from abc import ABC, abstractmethod


class CanvasSyncObserver(ABC):
    """
    A CanvasSyncObserver subscribes to events related to synchronisation from canvas
    """

    @abstractmethod
    def new_content_found(self, content: object): ...


class CanvasSyncObservable:
    """
    A CanvasSyncObservable publishes events related to synchronisation from canvas
    """

    def __init__(self):
        self._observers: list[CanvasSyncObserver] = []

    def notify_observers_for_updated_item(self, content: object):
        for observer in self._observers:
            observer.new_content_found(content)

    @property
    def observers(self) -> list[CanvasSyncObserver]:
        return self._observers



from datetime import datetime

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column

from libqcanvas.database.tables.creation_date import CreationDate
from libqcanvas.database.tables.id import ID


class CourseContentItem(ID, CreationDate):
    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))

    name: Mapped[str]
    body: Mapped[str | None]

    unlock_at: Mapped[datetime | None] = mapped_column(default=None)
    lock_at: Mapped[datetime | None] = mapped_column(default=None)
    # Tracks whether the page is locked
    can_view: Mapped[bool]



from sqlalchemy.orm import Mapped

from libqcanvas.database.tables import Resource

class WithResourcesOrnamentalMixin:
    resources: Mapped[list[Resource]]
    dead_resources: Mapped[list[Resource]]

class WithAttachmentsOrnamentalMixin:
    attachments: Mapped[list[Resource]]



from bs4 import Tag

import libqcanvas.database.tables as db
from libqcanvas.net.resources.extracting.link_extractor import LinkExtractor
from libqcanvas.net.resources.extracting.no_extractor_error import NoExtractorError
from libqcanvas.util import flatten


class Extractors:
    def __init__(self, *args: LinkExtractor):
        self._extractors = list(args)
        self._tag_whitelist = set(
            flatten([extractor.tag_whitelist for extractor in self._extractors])
        )

    def add(self, extractor: LinkExtractor):
        self._extractors.append(extractor)
        self._tag_whitelist.update(extractor.tag_whitelist)

    def extend(self, extractors: list[LinkExtractor]):
        self._extractors.extend(extractors)
        self._tag_whitelist.update(
            flatten([extractor.tag_whitelist for extractor in extractors])
        )

    def extractor_for_tag(self, tag: Tag) -> LinkExtractor:
        for extractor in self._extractors:
            if extractor.accepts_tag(tag):
                return extractor

        raise NoExtractorError()

    def extractor_for_resource(self, resource: db.Resource) -> LinkExtractor:
        split = str(resource.id).split(LinkExtractor.RESOURCE_ID_SEPARATOR, 2)
        extractor_name = split[0]

        for extractor in self._extractors:
            if extractor.extractor_id == extractor_name:
                return extractor

        raise NoExtractorError()

    @property
    def tag_whitelist(self) -> set[str]:
        return self._tag_whitelist



from .assignment_group_index import AssignmentGroup, AssignmentGroupIndexQueryData
from .assignment_group_index import DEFINITION as ASSIGNMENT_GROUP_INDEX_QUERY
from .assignment_index import (
    Assignment,
    AssignmentIndexQueryData,
    Submission,
    SubmissionComment,
    SubmissionCommentConnection,
    SubmissionConnection,
)
from .assignment_index import DEFINITION as ASSIGNMENT_INDEX_QUERY
from .attachment import Attachment
from .course_index import Course as ShallowCourse
from .course_index import CourseIndexQueryData, Term
from .course_index import DEFINITION as COURSE_INDEX_QUERY
from .course_mail import ConversationParticipant, CourseMailQueryData
from .course_mail import DEFINITION as COURSE_MAIL_QUERY
from .module_index import DEFINITION as MODULE_INDEX_QUERY
from .module_index import (
    File,
    Module,
    ModuleConnection,
    ModuleIndexQueryData,
    ModuleItem,
    Page,
)



import logging
from datetime import UTC, datetime

_logger = logging.getLogger(__name__)


def as_local(time: datetime | None) -> datetime | None:
    """
    Converts a foreign time to local time.
    If the time has no timezone, it is assumed to be in UTC. If it does have a timezone, it is converted appropriately.
    If the time is None, None is returned.
    :param time: The time to convert.
    :return: The time as local time, or None
    """
    if time is None:
        return None
    elif time.tzinfo is None:
        return time.replace(tzinfo=UTC).astimezone()
    else:
        return time.astimezone()



import logging
import platform
import re
import tempfile
from abc import ABC, abstractmethod
from asyncio import BoundedSemaphore, TaskGroup
from pathlib import Path, PurePath
from typing import Sequence

from aiofile import async_open
from asynctaskpool import AsyncTaskPool

import libqcanvas.database.tables as db
from libqcanvas.database import QCanvasDatabase
from libqcanvas.net.resources.extracting.extractors import Extractors
from libqcanvas.net.resources.extracting.link_extractor import LinkExtractor
from libqcanvas.net.resources.scanning.resource_scanner import ResourceScanner
from libqcanvas.task_master import register_reporter
from libqcanvas.task_master.reporters import ContextManagerReporter

_logger = logging.getLogger(__name__)
_temp_download_dir = Path(tempfile.gettempdir()) / "qcanvas"


class ResourceManager(ABC):
    def __init__(
        self,
        database: QCanvasDatabase,
        download_dest: Path,
        extractors: Extractors = Extractors(),
        max_concurrent_downloads: int = 8,
    ):
        super().__init__()
        self._download_pool = AsyncTaskPool[None]()
        self._db = database
        self._extractors = extractors
        self._scanner = ResourceScanner(extractors)
        self._download_dest = download_dest
        self._download_limiter = BoundedSemaphore(max_concurrent_downloads)

    def add_existing_resources(self, existing_resources: Sequence[db.Resource]):
        self._scanner.add_existing_resources(existing_resources)

    def is_video(self, resource: db.Resource) -> bool:
        # todo using the extractor to tell if its a video is kinda lacklustre
        return self._extractors.extractor_for_resource(resource).is_video_extractor

    async def batch_download(
        self,
        resources: list[db.Resource],
    ) -> None:
        downloads: list[tuple[LinkExtractor, db.Resource]] = []
        deferred_downloads: list[tuple[LinkExtractor, db.Resource]] = []

        for resource in resources:
            extractor = self._extractors.extractor_for_resource(resource)

            # Defer large downloads to do them 1 at a time
            if extractor.is_video_extractor:
                deferred_downloads.append((extractor, resource))
            else:
                downloads.append((extractor, resource))

        async with TaskGroup() as tg:
            for extractor, resource in downloads:
                tg.create_task(
                    self._start_download(extractor, resource, suppress_exceptions=True)
                )

        # Do large downloads one at a time. Parallelizing them is unlikely to speed anything up.
        for extractor, resource in deferred_downloads:
            await self._start_download(extractor, resource, suppress_exceptions=True)

    async def download(self, resource: db.Resource) -> None:
        if resource.download_state != db.ResourceDownloadState.DOWNLOADED:
            extractor = self._extractors.extractor_for_resource(resource)
            await self._start_download(extractor, resource)

    async def _start_download(
        self,
        extractor: LinkExtractor,
        resource: db.Resource,
        suppress_exceptions: bool = False,
    ) -> None:
        await self._download_pool.submit(
            resource.id,
            self._download_task(
                extractor, resource, suppress_exceptions=suppress_exceptions
            ),
        )

    async def _download_task(
        self,
        extractor: LinkExtractor,
        resource: db.Resource,
        *,
        suppress_exceptions: bool = False,
    ) -> None:
        download_destination = self.resource_download_location(resource)

        if download_destination.exists():
            _logger.info(
                f"Resource {resource.file_name} seems to already be downloaded, skipping download"
            )

            await self._finished_download(
                resource, final_file_size=download_destination.stat().st_size
            )
            return
        else:
            download_destination.parent.mkdir(parents=True, exist_ok=True)

        async with self._download_limiter:
            try:
                # Use a temporary destination to download the file to, then once it's done, move it to the proper location.
                # This way, if something goes wrong we won't have garbage files in the download directory.
                temp_destination = self._prepare_temp_download_dest(resource)

                with register_reporter(
                    ContextManagerReporter("Download", resource.file_name, 0)
                ) as prog:  # type: ContextManagerReporter
                    async for progress in extractor.download(
                        resource,
                        destination=temp_destination,
                    ):
                        prog.progress(progress.downloaded_bytes, progress.total_bytes)
                        self.on_download_progress(
                            resource=resource,
                            current=progress.downloaded_bytes,
                            total=progress.total_bytes,
                        )

                    # Copy the temp file to the proper location
                    await _async_copy(src=temp_destination, dest=download_destination)
            except Exception as e:
                await self._db.record_resource_download_failed(resource, message=str(e))
                self.on_download_failed(resource)

                if not suppress_exceptions:
                    raise RuntimeError() from e
                else:
                    _logger.warning("Download failed", exc_info=e)
            else:
                # Get the file size so we can update the DB record to be accurate
                file_size = download_destination.stat().st_size

                await self._finished_download(resource, final_file_size=file_size)
            finally:
                # Delete the temp file if it was created
                temp_destination.unlink(missing_ok=True)

    async def _finished_download(
        self, resource: db.Resource, final_file_size: int
    ) -> None:
        await self._db.record_resource_downloaded(resource, final_file_size)
        self.on_download_finished(resource)

    def _prepare_temp_download_dest(self, resource: db.Resource) -> Path:
        _temp_download_dir.mkdir(parents=True, exist_ok=True)
        temp_destination = _temp_download_dir / self.resource_download_file_name(
            resource
        )
        return temp_destination

    def resource_download_location(self, resource: db.Resource) -> Path:
        return (
            self._download_dest
            / self.course_folder_name(resource.course)
            / self.resource_download_file_name(resource)
        )

    def resource_download_file_name(self, resource: db.Resource) -> str:
        file_name, file_suffix = ResourceManager._split_resource_name(resource)
        file_suffix = f" [{resource.id}]{file_suffix}"
        # Ensure the filename is not too long, most filesystems permit up to 255 chars
        file_name = file_name[: 255 - len(file_suffix)]

        return self._replace_illegal_chars(file_name + file_suffix)

    def course_folder_name(self, course: db.Course) -> str:
        return self._replace_illegal_chars(course.name)

    @staticmethod
    def _split_resource_name(resource: db.Resource) -> tuple[str, str]:
        file = PurePath(resource.file_name)
        return file.stem, file.suffix

    @staticmethod
    def _replace_illegal_chars(file_name: str) -> str:
        if platform.system() == "Windows":
            return re.sub(r"[<>:\"/\\|?*]", "_", file_name)
        else:
            return file_name.replace("/", "_")

    @property
    def scanner(self) -> ResourceScanner:
        return self._scanner

    @property
    def extractors(self) -> Extractors:
        return self._extractors

    @property
    def downloads_folder(self) -> Path:
        return self._download_dest

    @abstractmethod
    def on_download_progress(
        self, resource: db.Resource, current: int, total: int
    ) -> None: ...

    @abstractmethod
    def on_download_failed(self, resource: db.Resource) -> None: ...

    @abstractmethod
    def on_download_finished(self, resource: db.Resource) -> None: ...


async def _async_copy(*, src: Path, dest: Path) -> None:
    async with async_open(src, "rb") as src_file, async_open(dest, "wb") as dest_file:
        async for chunk in src_file.iter_chunked(65535):
            await dest_file.write(chunk)



import logging
from typing import Sequence

from sqlalchemy.ext.asyncio import AsyncSession

import libqcanvas.database.tables as db
from libqcanvas.database import resource_links
from libqcanvas.net.resources.scanning.extracted_resources import ExtractedResources
from libqcanvas.net.sync.canvas_sync_observer import CanvasSyncObservable

_logger = logging.getLogger(__name__)


class ResourceLinker(CanvasSyncObservable):
    def __init__(self, session: AsyncSession):
        super().__init__()
        self._session = session

    async def link_resources(self, extracted_resource_links: list[ExtractedResources]):
        if len(extracted_resource_links) == 0:
            return

        await self._mark_content_links_as_dead(extracted_resource_links)
        await self._link_new_resources_and_restore_active_links(
            extracted_resource_links
        )

    async def _mark_content_links_as_dead(
        self, extracted_resource_links: list[ExtractedResources]
    ):
        await resource_links.bulk_change_embedded_link_states(
            self._session,
            [link.content for link in extracted_resource_links],
            link_state="residual",
        )

    async def _link_new_resources_and_restore_active_links(
        self, extracted_resource_links: list[ExtractedResources]
    ):
        for link in extracted_resource_links:
            existing_resource_links = await resource_links.get_associated_resource_ids(
                self._session, link.content, link_type="embedded"
            )
            newly_linked_resources: list[str] = []

            await self._create_links_for_resources(
                existing_resource_links=existing_resource_links,
                extracted_resource_links=link,
                newly_linked_resources=newly_linked_resources,
            )

            await self._create_links_for_invisible_resources(
                existing_resource_links=existing_resource_links,
                extracted_resource_links=link,
                newly_linked_resources=newly_linked_resources,
            )

    async def _create_links_for_resources(
        self,
        existing_resource_links: Sequence[str],
        extracted_resource_links: ExtractedResources,
        newly_linked_resources: list[str],
    ):
        for extracted_resource in extracted_resource_links.resources:
            if extracted_resource.id in existing_resource_links:
                # If a resource is still linked on this page, reactivate the link
                await resource_links.change_embedded_link_state(
                    session=self._session,
                    content_item=extracted_resource_links.content,
                    resource=extracted_resource,
                    link_state="active",
                )
            # Prevent adding duplicate links (duplicate links have been observed on some pages)
            elif extracted_resource.id not in newly_linked_resources:
                await resource_links.create_resource_link(
                    self._session,
                    content_item=extracted_resource_links.content,
                    resource=extracted_resource,
                )
                newly_linked_resources.append(extracted_resource.id)

                await self._add_resource_to_db_if_new(extracted_resource)

    async def _create_links_for_invisible_resources(
        self,
        existing_resource_links: Sequence[str],
        extracted_resource_links: ExtractedResources,
        newly_linked_resources: list[str],
    ):
        for resource in extracted_resource_links.invisible_resources:
            if (
                resource.id in existing_resource_links
                or resource.id in newly_linked_resources
            ):
                # If the resource was already on the page, or it has just been added, don't add it again
                continue

            await resource_links.create_resource_link(
                self._session,
                content_item=extracted_resource_links.content,
                resource=resource,
                link_state="residual",
            )
            newly_linked_resources.append(resource.id)
            await self._add_resource_to_db_if_new(resource)

    async def _add_resource_to_db_if_new(self, resource: db.Resource):
        does_not_exist = (await self._session.get(db.Resource, resource.id)) is None

        if does_not_exist:
            _logger.debug("New resource %s (id=%s)", resource.file_name, resource.id)
            self._session.add(resource)
            self.notify_observers_for_updated_item(resource)



from typing import Protocol, Sequence

from sqlalchemy.orm import Mapped

from libqcanvas.database.tables import CourseContentItem


class ContentGroup(Protocol):
    id: Mapped[str]
    name: Mapped[str]

    @property
    def items(self) -> Sequence[CourseContentItem]: ...



from bs4 import BeautifulSoup, Tag

NBSP = "\N{NO-BREAK SPACE}"


def remove_unwanted_whitespaces(contaminated_data: str | None) -> str | None:
    """
    Remove trailing spaces and replace non-breaking spaces
    :param contaminated_data: The string to clean
    :return: The sanitized string
    """
    if contaminated_data is None:
        return None

    return contaminated_data.replace(NBSP, " ").strip(f"\t {NBSP}")


def remove_screenreader_elements(contaminated_data: str | None) -> str | None:
    if contaminated_data is None:
        return None

    doc = BeautifulSoup(contaminated_data, "html.parser")

    for element in doc.find_all(class_="screenreader-only"):
        element.decompose()

    return str(doc)


def is_link_invisible(tag: Tag) -> bool:
    return tag.name == "a" and remove_unwanted_whitespaces(tag.text) == ""



from datetime import datetime
from typing import Sequence

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship

from libqcanvas.database.resource_links import (
    WithAttachmentsOrnamentalMixin,
    WithResourcesOrnamentalMixin,
    with_attachments,
    with_embedded_resources,
)
from libqcanvas.database.tables.base import Base
from libqcanvas.database.tables.course_content_item import CourseContentItem
from libqcanvas.database.tables.creation_date import CreationDate
from libqcanvas.database.tables.id import ID
from libqcanvas.database.tables.modification_date import ModificationDate
from libqcanvas.database.tables.module_page_type import ModulePageType
from libqcanvas.database.tables.resource import Resource


@with_attachments
class SubmissionComment(
    Base, ID, CreationDate, ModificationDate, WithAttachmentsOrnamentalMixin
):
    __tablename__ = "submission_comments"

    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))
    course: Mapped["Course"] = relationship()
    group_id: Mapped[str] = mapped_column(ForeignKey("assignment_groups.id"))
    group: Mapped["AssignmentGroup"] = relationship()
    assignment_id: Mapped[str] = mapped_column(ForeignKey("assignments.id"))
    assignment: Mapped["Assignment"] = relationship()
    submission_id: Mapped[str] = mapped_column(ForeignKey("assignment_submissions.id"))
    submission: Mapped["AssignmentSubmission"] = relationship(back_populates="comments")

    author: Mapped[str]
    body: Mapped[str]


@with_attachments
class AssignmentSubmission(
    Base, ID, CreationDate, ModificationDate, WithAttachmentsOrnamentalMixin
):
    __tablename__ = "assignment_submissions"

    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))
    course: Mapped["Course"] = relationship()
    group_id: Mapped[str] = mapped_column(ForeignKey("assignment_groups.id"))
    group: Mapped["AssignmentGroup"] = relationship()
    assignment_id: Mapped[str] = mapped_column(ForeignKey("assignments.id"))
    assignment: Mapped["Assignment"] = relationship(back_populates="submissions")

    comments: Mapped[list["SubmissionComment"]] = relationship(
        back_populates="submission", order_by=SubmissionComment.creation_date
    )

    attempt: Mapped[int]
    score: Mapped[float | None]


@with_embedded_resources
class Assignment(
    Base, CourseContentItem, ModificationDate, WithResourcesOrnamentalMixin
):
    __tablename__ = "assignments"

    group_id: Mapped[str] = mapped_column(ForeignKey("assignment_groups.id"))
    group: Mapped["AssignmentGroup"] = relationship(back_populates="assignments")

    submissions: Mapped[list["AssignmentSubmission"]] = relationship(
        back_populates="assignment", order_by=AssignmentSubmission.creation_date
    )

    due_date: Mapped[datetime | None]
    max_score: Mapped[float | None]
    position: Mapped[int]


class AssignmentGroup(Base, ID):
    __tablename__ = "assignment_groups"

    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))
    course: Mapped["Course"] = relationship(back_populates="assignment_groups")

    assignments: Mapped[list[Assignment]] = relationship(
        back_populates="group",
        order_by=Assignment.position,
        cascade="save-update, merge, delete",
    )

    name: Mapped[str]
    group_weight: Mapped[int]
    position: Mapped[int]

    @property
    def items(self) -> Sequence:
        return self.assignments


@with_attachments
@with_embedded_resources
class Message(
    Base,
    CourseContentItem,
    WithAttachmentsOrnamentalMixin,
    WithResourcesOrnamentalMixin,
):
    """
    Used for announcements and course mail
    """

    __tablename__ = "messages"

    sender_name: Mapped[str]
    has_been_read: Mapped[bool]
    course: Mapped["Course"] = relationship(back_populates="messages")


@with_embedded_resources
class Page(Base, CourseContentItem, ModificationDate, WithResourcesOrnamentalMixin):
    __tablename__ = "pages"

    module_id: Mapped[str] = mapped_column(ForeignKey("modules.id"))
    module: Mapped["Module"] = relationship(back_populates="pages")
    position: Mapped[int]
    page_type: Mapped[ModulePageType]


class Module(Base, ID):
    __tablename__ = "modules"

    name: Mapped[str]
    position: Mapped[int]
    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))
    course: Mapped["Course"] = relationship(back_populates="modules")

    pages: Mapped[list["Page"]] = relationship(
        back_populates="module",
        order_by=Page.position,
        cascade="save-update, merge, delete",
    )

    @property
    def items(self) -> Sequence[CourseContentItem]:
        return self.pages


class Course(Base, ID):
    __tablename__ = "courses"

    name: Mapped[str]
    panopto_folder_id: Mapped[str | None]
    term_id: Mapped[str] = mapped_column(ForeignKey("terms.id"))
    term: Mapped["Term"] = relationship(back_populates="courses")

    assignment_groups: Mapped[list[AssignmentGroup]] = relationship(
        back_populates="course",
        order_by=AssignmentGroup.position,
        cascade="save-update, merge, delete",
    )
    modules: Mapped[list["Module"]] = relationship(
        back_populates="course",
        order_by=Module.position,
        cascade="save-update, merge, delete",
    )
    messages: Mapped[list[Message]] = relationship(
        viewonly=True,
        order_by=Message.creation_date,
    )
    resources: Mapped[list[Resource]] = relationship(
        back_populates="course",
        order_by=Resource.discovery_date,
        cascade="save-update, merge, delete",
    )


class Term(Base, ID):
    __tablename__ = "terms"

    start_date: Mapped[datetime | None]
    end_date: Mapped[datetime | None]
    name: Mapped[str]

    courses: Mapped[list["Course"]] = relationship(
        back_populates="term",
        order_by=Course.name,
        cascade="save-update, merge, delete",
    )

    def __hash__(self):
        return (
            hash(self.id)
            ^ hash(self.start_date)
            ^ hash(self.end_date)
            ^ hash(self.name)
        )



from dataclasses import dataclass

import libqcanvas.database.tables as db


@dataclass(slots=True)
class ExtractedResources:
    content: db.AnyContentItem
    resources: list[db.Resource]
    invisible_resources: list[db.Resource]



import logging
from typing import Self

from libqcanvas.task_master.task_master import Reporter

_logger = logging.getLogger(__name__)


# todo: merge into base Reporter, not really any reason to have this functionality separate
class ContextManagerReporter(Reporter):
    """
    Report the progress of a task where the progress is a number with an upper limit

    Example:
    tm = MyTaskMaster()

    # Report the progress of some task with an incrementing number
    with tm.accept(ContextManagerReporter("My demo", "Count to 10", 10)) as prog:
        for i in range(0, 11):
            prog.progress(i)

    """

    def __init__(self, goal_name: str, step_name: str, total_work: int):
        super().__init__(goal_name, step_name)
        self._total = total_work

    def __enter__(self) -> Self:
        self.ensure_task_master_assigned()
        self.progress(0, self._total)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if exc_val is not None:
            self.failed(exc_val)



from .canvas_data_sanitiser import (
    is_link_invisible,
    remove_screenreader_elements,
    remove_unwanted_whitespaces,
)
from .collecting_task_group import CollectingTaskGroup
from .flatten import flatten
from .form_input_extractor import extract_form_inputs
from .timezone_util import as_local



from enum import Enum


class ResourceDownloadState(Enum):
    NOT_DOWNLOADED = 0
    DOWNLOADED = 1
    FAILED = 2

    @staticmethod
    def human_readable(value: "ResourceDownloadState"):
        match value:
            case ResourceDownloadState.NOT_DOWNLOADED:
                return "Not downloaded"
            case ResourceDownloadState.DOWNLOADED:
                return "Downloaded"
            case ResourceDownloadState.FAILED:
                return "Failed"

        raise ValueError(value)



from dataclasses import dataclass

from libqcanvas_clients.canvas import Announcement

from libqcanvas.net.canvas import CourseMailItem, PageWithContent
from libqcanvas.net.canvas.course_bundle import CourseBundle


@dataclass
class CanvasDataBundle:
    """
    A CanvasDataBundle is a collection of various data retrieved from canvas
    """

    courses: list[CourseBundle]
    pages: list[PageWithContent]
    messages: list[CourseMailItem | Announcement]
    course_panopto_folders: dict[str, str]



import logging
from asyncio import TaskGroup
from datetime import datetime

from libqcanvas_clients.canvas import CanvasClient
from sqlalchemy.ext.asyncio import AsyncSession

import libqcanvas.database.tables as db
import libqcanvas.gql_queries as gql
from libqcanvas.gql_queries import Module, ModuleItem, Page, ShallowCourse
from libqcanvas.net.canvas.course_bundle import CourseBundle
from libqcanvas.net.canvas.page_with_content import PageWithContent
from libqcanvas.net.constants import SYNC_GOAL
from libqcanvas.task_master import register_reporter
from libqcanvas.task_master.reporters import CompoundTaskReporter

_logger = logging.getLogger(__name__)


class CanvasPageLoader:
    def __init__(self, canvas_client: CanvasClient):
        self._canvas_client = canvas_client

    async def prepare_pages_for_update(
        self,
        session: AsyncSession,
        course_bundles: list[CourseBundle],
        last_update_time: datetime,
    ) -> list[PageWithContent]:
        pages = []

        for bundle in course_bundles:
            for module in bundle.modules:
                for position, module_item in enumerate(module.module_items):
                    if isinstance(module_item.content, gql.Page):
                        content = module_item.content
                        existing_page = await session.get(db.Page, content.q_id)
                        add_page = False

                        if existing_page is None:
                            # The page doesn't exist in our version of the database
                            add_page = True
                        elif not existing_page.can_view:
                            # The page is locked in our version of the database
                            add_page = True
                        elif self._is_module_item_out_of_date(
                            last_update_time, module_item
                        ):
                            add_page = True

                        if add_page:
                            _logger.debug(
                                "Preparing page %s (id=%s) for loading",
                                content.title,
                                content.q_id,
                            )
                            pages.append(
                                self._convert_to_page_with_content(
                                    page=module_item.content,
                                    module=module,
                                    course=bundle.course,
                                    position=position,
                                )
                            )
                        else:
                            _logger.debug(
                                "Discarding page %s (id=%s)",
                                content.title,
                                content.q_id,
                            )

        return pages

    async def load_pages_in_place(self, pages: list[PageWithContent]) -> None:
        if not pages:
            return

        with register_reporter(
            CompoundTaskReporter(SYNC_GOAL, "Load pages", len(pages))
        ) as prog:
            async with TaskGroup() as tg:
                for page in pages:
                    prog.attach(tg.create_task(self._load_page_content_inplace(page)))

    @staticmethod
    def _is_module_item_out_of_date(
        last_update_time: datetime, module_item: ModuleItem
    ) -> bool:
        return module_item.content.updated_at >= last_update_time

    @staticmethod
    def _convert_to_page_with_content(
        page: Page, module: Module, course: ShallowCourse, position: int
    ):
        return PageWithContent(
            q_id=page.q_id,
            name=page.title,
            updated_at=page.updated_at,
            created_at=page.created_at,
            module=module,
            course=course,
            position=position,
        )

    async def _load_page_content_inplace(self, page: PageWithContent):
        result = await self._canvas_client.get_page(
            page_id=page.q_id, course_id=page.course.q_id
        )
        _logger.debug('Loaded page %s (id="%s")', page.name, page.q_id)

        page.is_locked = result.locked_for_user
        page.content = result.body

        if result.lock_info is not None:
            page.unlock_at = result.lock_info.unlock_at
            page.lock_at = result.lock_info.lock_at



import logging
from pathlib import Path
from typing import Sequence

import sqlalchemy
from alembic import command
from alembic.config import Config
from alembic.runtime.migration import MigrationContext
from alembic.script import ScriptDirectory
from sqlalchemy import Connection, select, update
from sqlalchemy.ext.asyncio import async_sessionmaker, create_async_engine
from sqlalchemy.orm import selectin_polymorphic

import libqcanvas.database.tables as db
from libqcanvas.database.data_monolith import DataMonolith
from libqcanvas.database.tables.canvas_database_types import Base

_logger = logging.getLogger(__name__)


class QCanvasDatabase:
    def __init__(self, database_file: Path):
        self._database_file = database_file
        self._engine = create_async_engine(f"sqlite+aiosqlite:///{database_file}")
        self._session_maker = async_sessionmaker(self._engine, expire_on_commit=False)

    @staticmethod
    async def create(database_file: Path) -> "QCanvasDatabase":
        _database = QCanvasDatabase(database_file)
        await _database.init()
        return _database

    async def upgrade(self) -> None:
        if not self._database_file.exists():
            _logger.info("Database does not exist, not upgrading")
            return

        _logger.info("Preparing for database upgrade")

        config = self._get_alembic_config()
        script_dir = ScriptDirectory.from_config(config)

        async with self._engine.begin() as conn:

            def get_current_revision(connection: Connection) -> str:
                migration_context = MigrationContext.configure(connection=connection)
                return migration_context.get_current_revision()

            head = script_dir.get_revision("head").revision
            current_revision = await conn.run_sync(get_current_revision)

            _logger.debug("current head = %s, head = %s", current_revision, head)

            if current_revision == head:
                _logger.info("Upgrade not required")
                return
            else:
                # ScriptDirectory is 2hard4me... just use the command wrapper instead
                _logger.info("Upgrading database to %s", head)

                def run_upgrade(
                    connection: sqlalchemy.Connection, config: Config
                ) -> None:
                    config.attributes["connection"] = connection

                    command.upgrade(config, "head")

                await conn.run_sync(run_upgrade, config)

    async def init(self):
        is_first_run = not self._database_file.exists()

        if is_first_run:
            _logger.info("First run detected")

            async with self._engine.begin() as conn:

                def run_stamp(
                    connection: sqlalchemy.Connection, config: Config
                ) -> None:
                    config.attributes["connection"] = connection

                    command.stamp(config, "head")

                await conn.run_sync(Base.metadata.create_all)
                await conn.run_sync(run_stamp, self._get_alembic_config())
        else:
            _logger.info("Database already exists")

    @staticmethod
    def _get_alembic_config() -> Config:
        module_dir = Path(__file__).resolve().parents[1]
        alembic_dir = module_dir / "alembic"

        config = Config()
        config.set_main_option("script_location", str(alembic_dir.absolute()))

        return config

    def session(self):
        # Return type is missing intentionally.
        # Don't try and make this a context manager (which would normally be better) because pycharm still can't
        # determine the type correctly. Blech.
        return self._session_maker.begin()

    async def get_existing_course_ids(self) -> Sequence[str]:
        async with self._session_maker.begin() as session:
            return (await session.scalars(select(db.Course.id))).all()

    async def get_existing_resources(self) -> Sequence[db.Resource]:
        async with self._session_maker.begin() as session:
            stmt = select(db.Resource)
            return (await session.scalars(stmt)).all()

    async def get_resource(self, id: str) -> db.Resource:
        async with self._session_maker.begin() as session:
            stmt = (
                select(db.Resource)
                .where(db.Resource.id == id)
                .options(selectin_polymorphic(db.Resource, [db.PanoptoResource]))
            )
            return (await session.scalars(stmt)).one()

    async def record_resource_downloaded(
        self, resource: db.Resource, final_file_size: int
    ):
        await self._update_resource_download_state(
            resource,
            state=db.ResourceDownloadState.DOWNLOADED,
            final_file_size=final_file_size,
        )

    async def record_resource_download_failed(
        self, resource: db.Resource, message: str
    ):
        await self._update_resource_download_state(
            resource, message=message, state=db.ResourceDownloadState.FAILED
        )

    async def _update_resource_download_state(
        self,
        resource: db.Resource,
        *,
        state: db.ResourceDownloadState,
        final_file_size: int | None = None,
        message: str | None = None,
    ):
        async with self._session_maker.begin() as session:
            resource.download_state = state
            resource.download_error_message = message
            resource.file_size = final_file_size

            await session.execute(
                update(db.Resource),
                [
                    {
                        db.Resource.id.key: resource.id,
                        db.Resource.download_state.key: resource.download_state,
                        db.Resource.download_error_message.key: resource.download_error_message,
                        db.Resource.file_size.key: final_file_size,
                    }
                ],
            )

    async def get_data(self) -> DataMonolith:
        async with self.session() as session:
            return await DataMonolith.create(session)



from datetime import datetime

from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase, MappedAsDataclass

from libqcanvas.database.tables._tz_datetime import TZDateTime


class Base(DeclarativeBase, AsyncAttrs, MappedAsDataclass, init=False):
    type_annotation_map = {datetime: TZDateTime}



from typing import Sequence

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import joinedload, selectin_polymorphic, selectinload

import libqcanvas.database.tables as db


class DataMonolith:
    def __init__(
        self,
        courses: Sequence[db.Course],
        resources: Sequence[db.Resource],
        terms: Sequence[db.Term],
    ):
        self.courses = sorted(courses, key=lambda x: x.name)
        self.resources = self._create_resource_map(resources)
        self.terms = terms

    @staticmethod
    def _create_resource_map(resources):
        return {resource.id: resource for resource in resources}

    @staticmethod
    async def create(session: AsyncSession) -> "DataMonolith":
        eager_load_resources = [
            joinedload(db.Resource.course),
            selectin_polymorphic(db.Resource, [db.PanoptoResource]),
        ]

        query = select(db.Course).options(selectinload("*"))
        courses = (await session.scalars(query)).unique().all()
        query = select(db.Resource).options(*eager_load_resources)
        resources = (await session.scalars(query)).all()
        query = (
            select(db.Term)
            .order_by(db.Term.start_date.desc())
            .options(selectinload("*"))
        )
        terms = (await session.scalars(query)).unique().all()

        return DataMonolith(courses=courses, resources=resources, terms=terms)



import json
import logging
import re
from typing import Any, Optional

import httpx
from bs4 import BeautifulSoup, Tag
from httpx import URL, HTTPStatusError, Response
from libqcanvas_clients.canvas import CanvasClient
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
    wait_random,
)

from libqcanvas.net.panopto.iframe_panopto_video_info import IFramePanoptoVideoInfo
from libqcanvas.util import extract_form_inputs


class IFramePanoptoVideoExtractor:
    _logger = logging.getLogger(__name__)
    _info_json_pattern = re.compile(
        r"new PanoptoTS\.Embed\.EmbeddedViewer\((.+)\);[\r\n]"
    )

    def __init__(self, canvas_client: CanvasClient):
        self._canvas_client = canvas_client

    @retry(
        wait=wait_exponential(exp_base=1.2, max=10) + wait_random(0, 1),
        retry=retry_if_exception_type(HTTPStatusError),
        stop=stop_after_attempt(8),
    )
    async def retrieve_embedded_video_info(
        self, embedded_url: str | URL
    ) -> IFramePanoptoVideoInfo:
        # Panopto videos can be embedded on in 2 or 3 ways for some reason. One of those ways is through an "LTI"
        # (learning tool integration) and the way it works normally (on the browser) is that it makes several requests
        # which return pages with a form on them that is autosubmitted through javascript. This happens twice, as you
        # can see below.
        response = await self._make_request(embedded_url)
        response = await self._execute_next_form_action_from_response(response)
        response = await self._execute_next_form_action_from_response(response)

        response.raise_for_status()

        # self._logger.debug("Found video info at %s", response.url)

        return IFramePanoptoVideoInfo(**self._extract_info_json(response.text))

    async def _execute_next_form_action_from_response(
        self, response: Response
    ) -> Response:
        doc = BeautifulSoup(response.text, features="html.parser")
        return await self._execute_html_form_action(doc.find("form"))

    async def _execute_html_form_action(self, form_element: Tag) -> Response:
        attrs = form_element.attrs

        return await self._make_request(
            method=attrs["method"],
            url=attrs["action"],
            data=extract_form_inputs(form_element),
        )

    async def _make_request(
        self, url: str | URL, method: str = "GET", data: Optional[dict] = None
    ) -> Response:
        response = await self._canvas_client.make_generic_request(
            httpx.Request(url=url, method=method, data=data), follow_redirects=True
        )

        return response.raise_for_status()

    def _extract_info_json(self, response_text: str) -> dict[str, Any]:
        matches = self._info_json_pattern.findall(response_text)
        if len(matches) != 1:
            raise ValueError(
                f"Expected to have exactly 1 info json match, got {len(matches)}"
            )

        return json.loads(matches[0])



import logging
from collections import defaultdict
from typing import Literal, Protocol, Sequence, runtime_checkable

from sqlalchemy import Table, select, update
from sqlalchemy.dialects.sqlite import insert
from sqlalchemy.ext.asyncio import AsyncSession

from libqcanvas.database.tables.resource import Resource

_logger = logging.getLogger(__name__)


type LinkType = Literal["embedded", "attachment"]
type LinkState = Literal["residual", "active"]


@runtime_checkable
class HasEmbeddedResources(Protocol):
    __tablename__: str
    __resource_link_table__: Table
    id: object


@runtime_checkable
class HasAttachments(Protocol):
    __tablename__: str
    __attachment_link_table__: Table
    id: object


def _get_content_link_table(content_item: object | type, link_type: LinkType) -> Table:
    if link_type == "embedded":
        if isinstance(content_item, HasEmbeddedResources):
            return content_item.__resource_link_table__
        else:
            raise TypeError("content_item does not support embedded resources")
    elif link_type == "attachment":
        if isinstance(content_item, HasAttachments):
            return content_item.__attachment_link_table__
        else:
            raise TypeError("content_item does not support attachments")
    else:
        raise ValueError("link_type")


def _validate_link_type(link_type: str) -> None:
    if link_type not in ["embedded", "attachment"]:
        raise ValueError(f"link_type: {link_type}")


def _validate_link_state(link_state: str) -> None:
    if link_state not in ["active", "residual"]:
        raise ValueError(f"link_state: {link_state}")


async def create_resource_link(
    session: AsyncSession,
    content_item: object,
    resource: Resource,
    *,
    link_type: LinkType = "embedded",
    link_state: LinkState = "active",
) -> None:
    await create_resource_link_a(
        session,
        content_id=content_item.id,
        content_type=content_item,
        resource=resource,
        link_type=link_type,
        link_state=link_state,
    )


# todo what the hell am I supposed to call this?? i wish python had overloading
async def create_resource_link_a[T](
    session: AsyncSession,
    *,
    content_id: object,
    content_type: type[T] | object,
    resource: Resource,
    link_type: LinkType = "embedded",
    link_state: LinkState = "active",
) -> None:
    table = _get_content_link_table(content_type, link_type)
    query_params = dict(content_item_id=content_id, resource_id=resource.id)

    if link_type == "embedded":
        _validate_link_state(link_state)
        query_params["link_state"] = link_state
    elif link_type == "attachment":
        if link_state != "active":
            raise ValueError("attachments do not support link_state")

    stmt = (
        insert(table)
        .values(**query_params)
        .on_conflict_do_nothing(index_elements=["content_item_id", "resource_id"])
    )
    await session.execute(stmt)


async def get_associated_resource_ids(
    session: AsyncSession, content_item: object, *, link_type: LinkType = "embedded"
) -> Sequence[str]:
    table = _get_content_link_table(content_item, link_type)
    stmt = select(table.c.resource_id).where(table.c.content_item_id == content_item.id)
    return (await session.scalars(stmt)).all()


async def change_embedded_link_state(
    session: AsyncSession,
    content_item: object,
    resource: Resource,
    *,
    link_state: LinkState,
) -> None:
    if not isinstance(content_item, HasEmbeddedResources):
        raise TypeError("content_item")

    _validate_link_state(link_state)

    link_table = content_item.__resource_link_table__
    stmt = (
        update(link_table)
        .where(link_table.c.resource_id == resource.id)
        .values(link_state=link_state)
    )
    await session.execute(stmt)


async def bulk_change_embedded_link_states(
    session: AsyncSession,
    content_items: list[HasEmbeddedResources],
    *,
    link_state: LinkState,
) -> None:
    _validate_link_state(link_state)
    content_grouped_by_link_table: dict[Table, set] = defaultdict(set)

    for item in content_items:
        if not isinstance(item, HasEmbeddedResources):
            raise TypeError(item)

        content_grouped_by_link_table[item.__resource_link_table__].add(item.id)

    for link_table, content_ids in content_grouped_by_link_table.items():
        stmt = (
            update(link_table)
            .where(link_table.c.content_item_id.in_(content_ids))
            .values(link_state=link_state)
        )
        await session.execute(stmt)



import difflib
import logging
import re
from typing import Tuple

from libqcanvas_clients.panopto import FolderInfo, PanoptoClient

import libqcanvas.gql_queries as gql
from libqcanvas.net.panopto.folder_with_sessions import FolderWithSessions

_logger = logging.getLogger(__name__)


class PanoptoDataAggregator:
    _course_name_cleaner_pattern = re.compile(r"(.+?) \(")

    def __init__(self, panopto_client: PanoptoClient):
        self._panopto_client = panopto_client

    async def get_course_panopto_folders(
        self, courses: list[gql.ShallowCourse]
    ) -> dict[str, str]:
        """
        Identifies the panopto folder for each course specified.
        :param courses: The courses for which to find the panopto folder
        :return: Dict of course_id -> folder_id
        """
        _logger.info("Fetching panopto folder index")

        result = {}
        folders = await self.get_folders()
        course_map = {
            self._formatted_course_name(course): course.q_id for course in courses
        }

        for folder in folders:
            try:
                clean_name = self._clean_course_name(folder.name)
                closest_match = difflib.get_close_matches(
                    clean_name, course_map.keys(), 1
                )[0]
            except IndexError:
                _logger.debug(
                    'No folder "%s" has no close match. If this is a partial sync, this is expected.',
                    folder.name,
                )
                continue
            except ValueError:
                _logger.info('Folder "%s" has invalid name, ignoring', folder.name)
                continue

            course_id = course_map[closest_match]
            result[course_id] = folder.id

            _logger.debug(
                "Matched panopto folder '%s' (id='%s') to course id='%s'",
                folder.name,
                folder.id,
                course_id,
            )

        return result

    def _formatted_course_name(self, course: gql.ShallowCourse) -> str:
        clean_name = self._clean_course_name(course.name)
        return f"{course.term.name} - {clean_name}"

    def _clean_course_name(self, name: str) -> str:
        # Remove course codes, semester name and campus information from a course/folder name
        match = self._course_name_cleaner_pattern.search(name)

        if match is None:
            raise ValueError()

        return match.group(1)

    async def _get_lectures_task(
        self, folder: FolderInfo, course_id: str
    ) -> Tuple[str, FolderWithSessions]:
        sessions = await self._panopto_client.get_folder_sessions(folder.id)
        return course_id, FolderWithSessions(folder=folder, sessions=sessions.results)

    async def get_folders(self) -> list[FolderInfo]:
        folders = await self._panopto_client.get_folders()
        return self._discard_parent_folders(folders)

    @staticmethod
    def _discard_parent_folders(folders: list[FolderInfo]) -> list[FolderInfo]:
        def predicate(folder: FolderInfo) -> bool:
            return not folder.has_accessible_children

        return list(filter(predicate, folders))



import logging
from dataclasses import dataclass, field

from libqcanvas.gql_queries import Assignment, AssignmentGroup, Module, ShallowCourse

_logger = logging.getLogger(__name__)


@dataclass(slots=True)
class CourseBundle:
    course: ShallowCourse
    assignment_groups: list[AssignmentGroup] = field(default_factory=list)
    assignments: list[Assignment] = field(default_factory=list)
    modules: list[Module] = field(default_factory=list)



from datetime import datetime

from pydantic import BaseModel, Field


class SyncMeta(BaseModel):
    class Config:
        extra = "allow"

    last_sync_time: datetime = Field(
        default=datetime.fromisoformat("1900-01-01T00:00:00Z")
    )



from .decorators import with_attachments, with_embedded_resources
from .ornamental_mixins import (
    WithAttachmentsOrnamentalMixin,
    WithResourcesOrnamentalMixin,
)
from .resource_link_manipulators import (
    bulk_change_embedded_link_states,
    change_embedded_link_state,
    create_resource_link,
    create_resource_link_a,
    get_associated_resource_ids,
)



import asyncio
import logging
from datetime import datetime
from typing import Sequence

from gql import gql
from libqcanvas_clients.canvas import CanvasClient

from libqcanvas import gql_queries
from libqcanvas.gql_queries import (
    Assignment,
    AssignmentGroup,
    AssignmentGroupIndexQueryData,
    AssignmentIndexQueryData,
    ConversationParticipant,
    CourseIndexQueryData,
    CourseMailQueryData,
    Module,
    ModuleIndexQueryData,
    ShallowCourse,
    Term,
)
from libqcanvas.net.canvas.course_bundle import CourseBundle
from libqcanvas.net.canvas.course_mail_item import CourseMailItem
from libqcanvas.net.constants import SYNC_GOAL
from libqcanvas.task_master import register_reporter
from libqcanvas.task_master.reporters import AtomicTaskReporter, CompoundTaskReporter
from libqcanvas.util import CollectingTaskGroup, flatten

_logger = logging.getLogger(__name__)


class CanvasDataAggregator:
    def __init__(self, canvas_client: CanvasClient):
        self._canvas_client = canvas_client

    async def pull_user_id(self) -> str:
        return await self._canvas_client.get_current_user_id()

    async def get_all_course_mail(self, current_user_id: str) -> list[CourseMailItem]:
        _logger.info("Fetching all course mail")

        with register_reporter(AtomicTaskReporter(SYNC_GOAL, "Fetch course mail")):
            result = await self._gql_query(
                gql_queries.COURSE_MAIL_QUERY, user_id=current_user_id
            )

        raw_mail_data = CourseMailQueryData(
            **result
        ).legacy_node.conversations_connection.nodes

        return self._convert_mail(raw_mail_data)

    @staticmethod
    def _convert_mail(
        raw_mail_data: list[ConversationParticipant],
    ):
        return [CourseMailItem.from_query_result(mail) for mail in raw_mail_data]

    async def pull_courses(
        self, already_indexed_course_ids: Sequence[str], include_old_courses: bool
    ) -> list[CourseBundle]:
        def course_is_accessible(course: ShallowCourse) -> bool:
            if course.settings is None:
                _logger.info(
                    f"Course '{course.name}' (id={course.q_id}) has settings=None, assuming it is inaccessible and ignoring it."
                )
                return False
            else:
                return True

        def course_belongs_to_term(course: ShallowCourse, term: Term) -> bool:
            return course.term.q_id == term.q_id

        def course_term_has_no_dates(course: ShallowCourse) -> bool:
            if course.term:
                if not course.term.start_at and not course.term.end_at:
                    return True

            return False

        def course_is_old_but_unindexed(course: ShallowCourse) -> bool:
            nonlocal already_indexed_course_ids
            return course.q_id not in already_indexed_course_ids

        shallow_courses = list(
            filter(course_is_accessible, await self._get_shallow_course_data())
        )

        if len(shallow_courses) == 0:
            return []

        latest_term = self._find_latest_term(shallow_courses)

        courses_to_pull = [
            course
            for course in shallow_courses
            if course_belongs_to_term(course, latest_term)
            or course_term_has_no_dates(course)
            or (include_old_courses and course_is_old_but_unindexed(course))
        ]

        return await self._pull_detailed_courses_by_id(courses_to_pull)

    async def _get_shallow_course_data(self) -> list[ShallowCourse]:
        with register_reporter(AtomicTaskReporter(SYNC_GOAL, "Fetch course indexes")):
            result = await self._gql_query(gql_queries.COURSE_INDEX_QUERY)

        return CourseIndexQueryData(**result).all_courses

    async def _pull_detailed_courses_by_id(
        self, courses: list[ShallowCourse]
    ) -> list[CourseBundle]:
        if _logger.isEnabledFor(logging.INFO):
            _logger.info("Fetching index for courses:")
            for course in courses:
                _logger.info(f"> '{course.name}' (id={course.q_id})")

        if len(courses) == 0:
            return []

        with register_reporter(
            CompoundTaskReporter(SYNC_GOAL, "Fetch course data", len(courses))
        ) as prog:
            async with CollectingTaskGroup[CourseBundle]() as tg:
                for course in courses:
                    prog.attach(tg.create_task(self._fetch_course(course)))

        return tg.results

    async def _fetch_course(self, course: ShallowCourse) -> CourseBundle:
        # todo progress reporting could be a bit more fine grained here

        bundle = CourseBundle(course=course)

        async def assignments_task():
            nonlocal bundle
            bundle.assignment_groups = await self._fetch_assignment_groups(course.q_id)

            async with CollectingTaskGroup[list[Assignment]]() as tg:
                for assignment_group in bundle.assignment_groups:
                    tg.create_task(self._fetch_assignments(assignment_group.q_id))

            bundle.assignments = flatten(tg.results)

        async def modules_task():
            nonlocal bundle
            bundle.modules = await self._fetch_modules(course.q_id)

        await asyncio.gather(assignments_task(), modules_task())

        return bundle

    async def _fetch_assignment_groups(self, course_id: str) -> list[AssignmentGroup]:
        query_result = await self._gql_query(
            gql_queries.ASSIGNMENT_GROUP_INDEX_QUERY,
            course_id=course_id,
        )
        parsed_result = AssignmentGroupIndexQueryData(**query_result)

        if not parsed_result.legacy_node:
            return []

        return parsed_result.legacy_node.assignment_groups or []

    async def _fetch_assignments(self, group_id: str) -> list[Assignment]:
        query_result = await self._gql_query(
            gql_queries.ASSIGNMENT_INDEX_QUERY, group_id=group_id
        )
        parsed_result = AssignmentIndexQueryData(**query_result)

        if (
            not parsed_result
            or not parsed_result.node
            or not parsed_result.node.assignments_connection
            or not parsed_result.node.assignments_connection.nodes
        ):
            return []

        return list(filter(None, parsed_result.node.assignments_connection.nodes))

    async def _fetch_modules(self, course_id: str) -> list[Module]:
        query_result = await self._gql_query(
            gql_queries.MODULE_INDEX_QUERY, course_id=course_id
        )
        parsed_result = ModuleIndexQueryData(**query_result)

        if (
            not parsed_result.legacy_node
            or not parsed_result.legacy_node.modules_connection
            or not parsed_result.legacy_node.modules_connection.nodes
        ):
            return []

        return list(filter(None, parsed_result.legacy_node.modules_connection.nodes))

    async def _gql_query(self, query: str, **kwargs):
        return await self._canvas_client.graphql_query(gql(query), kwargs)

    @staticmethod
    def _find_latest_term(courses: list[ShallowCourse]) -> Term:
        terms = [course.term for course in courses if course.term is not None]
        # For terms that don't have an end date (for whatever reason), just put them at the end of the list.
        # Don't use datetime.min because it can cause errors when doing timezone maths
        default_end_at = datetime.fromisoformat("1900-01-01T00:00:00Z")
        sorted_terms = sorted(terms, key=lambda term: term.end_at or default_end_at)

        return sorted_terms[-1]



import logging
from abc import ABC, abstractmethod

from libqcanvas.task_master.task_id import TaskID

_logger = logging.getLogger(__name__)


class Reporter:
    """
    A reporter reports the progress of a task to a task master
    """

    def __init__(self, goal_name: str, step_name: str):
        self.task_master: TaskMaster | None = None
        self.task_id = TaskID(goal_name, step_name)

    def ensure_task_master_assigned(self) -> None:
        if self.task_master is None:
            raise Exception("Reporter has not been assigned to any task master yet")

    def failed(self, context: str | Exception) -> None:
        self.ensure_task_master_assigned()
        self.task_master.report_failed(self.task_id, context)

    def progress(self, current: int, total: int) -> None:
        self.ensure_task_master_assigned()
        assert current <= total
        self.task_master.report_progress(self.task_id, current, total)


class TaskMaster(ABC):
    """
    A task master is the central entity that task progress is reported to from reporters
    """

    @abstractmethod
    def report_failed(self, task_id: TaskID, context: str | Exception) -> None: ...

    @abstractmethod
    def report_progress(self, task_id: TaskID, current: int, total: int) -> None:
        """
        Implementations can expect:
        - TaskID for a specific task to always be the same
        - For intermediate tasks to be reported as (current=0, total=0)
        - For total to change as the task progresses
        - For finished tasks to be reported as current = total (unless total = 0, in which case it is an intermediate task)
        :param task_id:
        :param current:
        :param total:
        :return:
        """

    def accept[T: Reporter](self, reporter: T) -> T:
        reporter.task_master = self
        return reporter



from dataclasses import dataclass
from datetime import datetime

from libqcanvas.gql_queries import ConversationParticipant


@dataclass
class CourseMailItem:
    id: str
    subject: str
    body: str
    author_name: str
    course_name: str
    course_id: str
    date: datetime

    @staticmethod
    def from_query_result(data: ConversationParticipant) -> "CourseMailItem":
        outer_data = data.conversation
        message_data = outer_data.conversation_messages_connection.nodes[0]

        return CourseMailItem(
            id=message_data.q_id,
            body=message_data.body,
            author_name=message_data.author.name,
            course_name=outer_data.context_name,
            subject=outer_data.subject,
            course_id=outer_data.context_id,
            date=message_data.created_at,
        )



import logging
from typing import NamedTuple, Optional, Sequence

import bs4
from asynctaskpool import AsyncTaskPool
from asynctaskpool.task_failed_exception import TaskFailedError
from bs4 import Tag

import libqcanvas.database.tables as db
from libqcanvas.net.constants import SYNC_GOAL
from libqcanvas.net.resources.extracting.extractors import Extractors
from libqcanvas.net.resources.extracting.no_extractor_error import NoExtractorError
from libqcanvas.net.resources.scanning.extracted_resources import ExtractedResources
from libqcanvas.task_master import register_reporter
from libqcanvas.task_master.reporters import CompoundTaskReporter
from libqcanvas.util import CollectingTaskGroup, is_link_invisible

_logger = logging.getLogger(__name__)


class _HiddenResource(NamedTuple):
    """
    A HiddenResource is a resource that, due to stupid shit that canvas does, may be in the HTML of a course page but
    are completely invisible, due to containing no text. I suspect these are created from something weird the canvas
    RCE (rich content editor) is doing when people edit links/pages.

    See Also
    --------
    libqcanvas.util.canvas_data_sanitiser.is_link_invisible
    """

    resource: db.Resource
    is_hidden: bool


class ResourceScanner:
    """
    A ResourceScanner uses Extractors to extract resources from course content
    """

    def __init__(self, extractors: Extractors):
        self._lazy_resource_cache = AsyncTaskPool[db.Resource]()
        self._extractor_collection = extractors

    def add_existing_resources(self, existing_resources: Sequence[db.Resource]):
        resources_mapped_by_id = {
            resource.id: resource for resource in existing_resources
        }
        self._lazy_resource_cache.update_results(resources_mapped_by_id)

    async def scan_content_for_resources(
        self, content_items: Sequence[db.AnyContentItem]
    ) -> list[ExtractedResources]:
        _logger.info("Scanning %i content items for resources", len(content_items))

        if len(content_items) == 0:
            return []

        with register_reporter(
            CompoundTaskReporter(SYNC_GOAL, "Scan for resources", len(content_items))
        ) as prog:
            async with CollectingTaskGroup[ExtractedResources]() as tg:
                for content_item in content_items:
                    prog.attach(
                        tg.create_task(self._extract_content_resources(content_item))
                    )

        return tg.results

    async def _extract_content_resources(
        self, content: db.AnyContentItem
    ) -> ExtractedResources:
        async with CollectingTaskGroup[_HiddenResource | None]() as tg:
            for tag in self._extract_tags_from_page(content):
                tg.create_task(
                    self._extract_resource(
                        tag=tag, course_id=content.course_id, content_id=content.id
                    )
                )

        invisible_resources = []
        visible_resources = []

        for resource, is_invisible in filter(None, tg.results):
            if is_invisible:
                invisible_resources.append(resource)
            else:
                visible_resources.append(resource)

        _logger.debug(
            "Found %i resources (%i invisible) on %s %s (id='%s')",
            len(invisible_resources) + len(visible_resources),
            len(invisible_resources),
            type(content).__name__,
            content.name,
            content.id,
        )

        return ExtractedResources(
            content=content,
            resources=visible_resources,
            invisible_resources=invisible_resources,
        )

    async def _extract_resource(
        self, tag: Tag, course_id: str, content_id: str
    ) -> Optional[_HiddenResource]:
        file_id = None

        try:
            extractor = self._extractor_collection.extractor_for_tag(tag)
            file_id = extractor.resource_id_from_tag(tag)

            result = await self._lazy_resource_cache.submit(
                task_id=file_id,
                future=extractor.resource_from_tag(
                    tag, course_id=course_id, resource_id=file_id
                ),
            )

            return _HiddenResource(resource=result, is_hidden=is_link_invisible(tag))
        except TaskFailedError as e:
            _logger.warning(
                "Extraction failed for file_id=%s on page id=%s",
                file_id or "(no id)",
                content_id,
                exc_info=e,
            )
        except NoExtractorError:
            pass
        except Exception as e:
            _logger.warning("Could not extract resource", exc_info=e)
            pass

    def _extract_tags_from_page(self, page: db.AnyContentItem) -> list[Tag]:
        if page.body is None:
            return []

        doc = bs4.BeautifulSoup(page.body, "html.parser")
        return doc.find_all(self._extractor_collection.tag_whitelist)



import libqcanvas.database.tables as db
from libqcanvas.net.sync.canvas_sync_observer import CanvasSyncObserver


class NewContentCollector(CanvasSyncObserver):
    def __init__(self):
        super().__init__()
        self._new_content: list[db.AnyContentItem] = []

    def new_content_found(self, content: object):
        if isinstance(content, db.CourseContentItem):
            self._new_content.append(content)

    @property
    def new_content(self) -> list[db.AnyContentItem]:
        return self._new_content



from operator import and_

from sqlalchemy import Column, ForeignKey, String, Table
from sqlalchemy.orm import relationship

from libqcanvas.database.tables.base import Base
from libqcanvas.database.tables.resource import Resource


def with_embedded_resources[T](cls: type[T]) -> type[T]:
    if hasattr(cls, "__tablename__"):
        tablename: str = cls.__tablename__
        link_table = Table(
            f"{tablename}_resource_links",
            Base.metadata,
            Column(
                "content_item_id",
                ForeignKey(f"{tablename}.id"),
                primary_key=True,
            ),
            Column("resource_id", ForeignKey("resources.id"), primary_key=True),
            Column("link_state", String),
        )
        cls.__resource_link_table__ = link_table
        cls.resources = relationship(
            Resource,
            secondary=link_table,
            primaryjoin=and_(
                link_table.c.content_item_id == cls.id,
                link_table.c.link_state == "active",
            ),
            order_by=Resource.id,
            overlaps="dead_resources",
            viewonly=True,
        )
        cls.dead_resources = relationship(
            Resource,
            secondary=link_table,
            primaryjoin=and_(
                link_table.c.content_item_id == cls.id,
                link_table.c.link_state == "residual",
            ),
            order_by=Resource.id,
            overlaps="resources",
            viewonly=True,
        )
    else:
        raise RuntimeError()

    return cls


def with_attachments[T](cls: type[T]) -> type[T]:
    if hasattr(cls, "__tablename__"):
        tablename: str = cls.__tablename__
        link_table = Table(
            f"{tablename}_attachment_links",
            Base.metadata,
            Column(
                "content_item_id",
                ForeignKey(f"{tablename}.id"),
                primary_key=True,
            ),
            Column("resource_id", ForeignKey("resources.id"), primary_key=True),
        )
        cls.__attachment_link_table__ = link_table
        cls.attachments = relationship(
            Resource,
            secondary=link_table,
            primaryjoin=link_table.c.content_item_id == cls.id,
            order_by=Resource.id,
            viewonly=True,
        )
    else:
        raise RuntimeError()

    return cls



"""Initial revision

Revision ID: 3c4e800178ab
Revises:
Create Date: 2024-11-29 16:45:55.161495

"""

from typing import Sequence, Union

# revision identifiers, used by Alembic.
revision: str = "3c4e800178ab"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    pass


def downgrade() -> None:
    pass



"""
Generated by qenerate plugin=pydantic_v2. DO NOT MODIFY MANUALLY!
"""

from collections.abc import Callable  # noqa: F401 # pylint: disable=W0611
from datetime import datetime  # noqa: F401 # pylint: disable=W0611
from enum import Enum  # noqa: F401 # pylint: disable=W0611
from typing import Any, Optional, Union  # noqa: F401 # pylint: disable=W0611

from pydantic import (
    BaseModel,
    # noqa: F401 # pylint: disable=W0611
    Field,
)

DEFINITION = """
query AssignmentGroupIndex($course_id: ID!) {
    legacyNode(_id: $course_id, type: Course) {
        ... on Course {
            assignmentGroups {
                id
                legacy_id: _id
                name
                groupWeight
            }
        }
    }
}
"""


class ConfiguredBaseModel(BaseModel):
    class Config:
        extra = "forbid"


class Node(ConfiguredBaseModel): ...


class AssignmentGroup(ConfiguredBaseModel):
    q_id: str = Field(..., alias="id")
    legacy_id: str = Field(..., alias="legacy_id")
    name: Optional[str] = Field(..., alias="name")
    group_weight: Optional[float] = Field(..., alias="groupWeight")


class Course(Node):
    assignment_groups: Optional[list[AssignmentGroup]] = Field(
        ..., alias="assignmentGroups"
    )


class AssignmentGroupIndexQueryData(ConfiguredBaseModel):
    legacy_node: Optional[Union[Course, Node]] = Field(..., alias="legacyNode")


def query(query_func: Callable, **kwargs: Any) -> AssignmentGroupIndexQueryData:
    """
    This is a convenience function which queries and parses the data into
    concrete types. It should be compatible with most GQL clients.
    You do not have to use it to consume the generated data classes.
    Alternatively, you can also mime and alternate the behavior
    of this function in the caller.

    Parameters:
        query_func (Callable): Function which queries your GQL Server
        kwargs: optional arguments that will be passed to the query function

    Returns:
        AssignmentGroupIndexQueryData: queried data parsed into generated classes
    """
    raw_data: dict[Any, Any] = query_func(DEFINITION, **kwargs)
    return AssignmentGroupIndexQueryData(**raw_data)



import logging
from asyncio import BoundedSemaphore
from pathlib import Path
from typing import Optional, Type

from libqcanvas_clients.canvas import CanvasClient, CanvasClientConfig
from libqcanvas_clients.panopto import (
    FakePanoptoClient,
    PanoptoClient,
    PanoptoClientConfig,
)

import libqcanvas.database.tables as db
from libqcanvas.database import QCanvasDatabase
from libqcanvas.database.data_monolith import DataMonolith
from libqcanvas.net.resources.download.resource_manager import ResourceManager
from libqcanvas.net.resources.extracting import (
    CanvasFileExtractor,
    CanvasMediaObjectExtractor,
    EmbeddedPanoptoExtractor,
)
from libqcanvas.net.resources.extracting.extractors import Extractors
from libqcanvas.net.resources.extracting.link_extractor import LinkExtractor
from libqcanvas.net.sync.sync_receipt import SyncReceipt
from libqcanvas.net.sync.synchronization_client import SynchronizationClient
from libqcanvas.task_master import register_reporter
from libqcanvas.task_master.reporters import AtomicTaskReporter

_logger = logging.getLogger(__name__)


class QCanvas[U: ResourceManager]:
    def __init__(
        self,
        canvas_config: CanvasClientConfig,
        panopto_config: Optional[PanoptoClientConfig],
        storage_path: Path,
        resource_manager_class: Type[U],
    ):
        storage_path.mkdir(parents=True, exist_ok=True)

        self._data_cache = None
        self._cache_semaphore = BoundedSemaphore()
        self.canvas_client = CanvasClient(client_config=canvas_config)

        if panopto_config is None:
            self.panopto_client = FakePanoptoClient()
        else:
            self.panopto_client = PanoptoClient(
                client_config=panopto_config, canvas_client=self.canvas_client
            )

        self.database = QCanvasDatabase(storage_path / "qcanvas.v2.db")
        self.resource_manager: U = self._create_resource_manager(
            manager=resource_manager_class,
            database=self.database,
            download_dest=storage_path / "downloads",
            extractors=Extractors(*self._create_extractors()),
        )
        self._synchronization_client = SynchronizationClient(
            database=self.database,
            canvas_client=self.canvas_client,
            panopto_client=self.panopto_client,
            resource_scanner=self.resource_manager.scanner,
            sync_meta_store_location=storage_path / "sync_meta.json",
        )
        self._init_called = False

    def _create_extractors(self) -> list[LinkExtractor]:
        result = [
            CanvasFileExtractor(self.canvas_client),
            CanvasMediaObjectExtractor(self.canvas_client),
        ]

        if not isinstance(self.panopto_client, FakePanoptoClient):
            result.append(
                EmbeddedPanoptoExtractor(self.canvas_client, self.panopto_client)
            )

        return result

    @staticmethod
    def _create_resource_manager(
        manager: Type[U],
        database: QCanvasDatabase,
        download_dest: Path,
        extractors: Extractors,
    ) -> ResourceManager:
        return manager(
            database=database, download_dest=download_dest, extractors=extractors
        )

    async def init(self) -> None:
        await self.database.init()
        self.resource_manager.add_existing_resources(
            await self.database.get_existing_resources()
        )

        self._init_called = True

    async def synchronise_canvas(self, quick_sync: bool = False) -> SyncReceipt:
        assert self._init_called
        _logger.info("Synchronising")

        receipt = await self._synchronization_client.synchronise_canvas(
            include_old_courses=quick_sync
        )
        await self._reload_data()
        return receipt

    async def _reload_data(self) -> None:
        self._data_cache = None
        await self.load()

    async def download(self, resource: db.Resource) -> None:
        assert self._init_called
        _logger.info(
            "Downloading %s (course_id=%s)", resource.file_name, resource.course_id
        )

        await self.resource_manager.download(resource)

    async def load(self) -> DataMonolith:
        assert self._init_called

        async with self._cache_semaphore:
            if self._data_cache is None:
                _logger.info("Loading all data from database")

                # Reading the entire DB can take a few seconds (especially when in portable mode on a shitty usb stick)
                with register_reporter(AtomicTaskReporter("Load", "Read database")):
                    self._data_cache = await self.database.get_data()

            return self._data_cache



import asyncio
from asyncio import Task


class CollectingTaskGroup[T](asyncio.TaskGroup):
    def __init__(self):
        super().__init__()
        self._preserved_tasks = []
        self._results = []

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await super().__aexit__(exc_type, exc_val, exc_tb)

        self._results = [task.result() for task in self._preserved_tasks if task.done()]
        self._preserved_tasks.clear()

    def create_task(self, coro, *, name=None, context=None) -> Task:
        result = super().create_task(coro, name=name, context=context)
        self._preserved_tasks.append(result)
        return result

    @property
    def results(self) -> list[T]:
        return self._results



from datetime import datetime

from sqlalchemy import Text, TypeDecorator


class TZDateTime(TypeDecorator):
    impl = Text
    cache_ok = True

    def process_bind_param(self, value: datetime, dialect):
        if value is not None:
            if not value.tzinfo or value.tzinfo.utcoffset(value) is None:
                raise TypeError("tzinfo is required")
            value = value.isoformat()
        return value

    def process_result_value(self, value: str, dialect):
        if value is not None:
            value = datetime.fromisoformat(value)
        return value



"""
WithResourcesOrnamentalMixin and WithAttachmentsOrnamentalMixin are stubs that tell pycharm about the extra attributes added by the resource link decorators
"""


class WithResourcesOrnamentalMixin:
    pass


class WithAttachmentsOrnamentalMixin:
    pass



from .canvas_database_types import (
    Assignment,
    AssignmentGroup,
    AssignmentSubmission,
    Course,
    CourseContentItem,
    Message,
    ModificationDate,
    Module,
    Page,
    SubmissionComment,
    Term,
)
from .content_group import ContentGroup
from .module_page_type import ModulePageType
from .resource import PanoptoResource, Resource
from .resource_download_state import ResourceDownloadState
from .resource_link_state import ResourceLinkState

type AnyContentItem = Assignment | Message | Page



"""
Generated by qenerate plugin=pydantic_v2. DO NOT MODIFY MANUALLY!
"""

from collections.abc import Callable  # noqa: F401 # pylint: disable=W0611
from datetime import datetime  # noqa: F401 # pylint: disable=W0611
from enum import Enum  # noqa: F401 # pylint: disable=W0611
from typing import Any, Optional, Union  # noqa: F401 # pylint: disable=W0611

from pydantic import (
    BaseModel,
    # noqa: F401 # pylint: disable=W0611
    Field,
)

DEFINITION = """
fragment Attachment on File {
    id
    url
    displayName
    mimeClass
    contentType
    createdAt
    updatedAt
}
"""


class ConfiguredBaseModel(BaseModel):
    class Config:
        extra = "forbid"


class Attachment(ConfiguredBaseModel):
    q_id: str = Field(..., alias="id")
    url: Optional[str] = Field(..., alias="url")
    display_name: Optional[str] = Field(..., alias="displayName")
    mime_class: Optional[str] = Field(..., alias="mimeClass")
    content_type: Optional[str] = Field(..., alias="contentType")
    created_at: Optional[datetime] = Field(..., alias="createdAt")
    updated_at: Optional[datetime] = Field(..., alias="updatedAt")



import logging

from libqcanvas_clients.canvas import Announcement

import libqcanvas.gql_queries as gql
from libqcanvas.net.canvas import CourseMailItem, PageWithContent
from libqcanvas.net.canvas.course_bundle import CourseBundle
from libqcanvas.net.constants import SYNC_GOAL
from libqcanvas.net.sync._canvas_data_bundle import CanvasDataBundle
from libqcanvas.net.sync.api_data_importer import APIDataImporter
from libqcanvas.task_master import register_reporter
from libqcanvas.task_master.reporters import AtomicTaskReporter

_logger = logging.getLogger(__name__)


class DataAdapter:
    """
    A DataAdapter deeply iterates over everything in a CanvasDataBundle and adds this data to the database using an APIDataImporter
    """

    def __init__(self, importer: APIDataImporter):
        self._converter = importer

    async def convert_and_add_to_database(self, canvas_sync_data: CanvasDataBundle):
        with register_reporter(AtomicTaskReporter(SYNC_GOAL, "Add data to database")):
            await self._store_partial_course_data(
                bundles=canvas_sync_data.courses,
                course_panopto_folders=canvas_sync_data.course_panopto_folders,
            )
            await self._store_pages(canvas_sync_data.pages)
            await self._store_messages(
                canvas_sync_data.messages,
                known_course_ids=[
                    bundle.course.q_id for bundle in canvas_sync_data.courses
                ],
            )

    async def _store_partial_course_data(
        self, bundles: list[CourseBundle], course_panopto_folders: dict[str, str]
    ):
        for term in self._flatten_terms(bundles):
            await self._converter.convert_and_store_term(term)

        for bundle in bundles:
            course_id = bundle.course.q_id
            await self._converter.convert_and_store_course(
                bundle.course,
                panopto_folder_id=course_panopto_folders.get(course_id, None),
            )

            for position, module in enumerate(bundle.modules):
                await self._converter.convert_and_store_module(
                    module, course_id=course_id, position=position
                )

            legacy_id_to_qgl_id: dict[str, str] = {}

            for position, assignment_group in enumerate(bundle.assignment_groups):
                # For some reason, assignment_group_id is a legacy ID despite being part of the newer graphql system.
                # Since assignment objects don't have the newer ID available, we need to store what it maps to.
                # An alternative to this would be to just use the legacy ID for assignment groups, but I would rather
                # just use the new ids where possible.
                legacy_id_to_qgl_id[assignment_group.legacy_id] = assignment_group.q_id

                await self._converter.convert_and_store_assignment_group(
                    assignment_group,
                    course_id=course_id,
                    position=position,
                )

            for assignment in bundle.assignments:
                group_id = legacy_id_to_qgl_id[assignment.assignment_group_id]

                await self._converter.convert_and_store_assignment(
                    assignment, group_id=group_id
                )

                for submission in assignment.submissions_connection.nodes:
                    await self._converter.convert_and_store_assignment_submission(
                        submission,
                        assignment_id=assignment.q_id,
                        course_id=course_id,
                        group_id=group_id,
                    )

                    for attachment in submission.attachments:
                        await self._converter.convert_and_store_attachment(
                            attachment,
                            course_id=course_id,
                            parent_api_object=submission,
                        )

                    for comment in submission.comments_connection.nodes:
                        await self._converter.convert_and_store_submission_comment(
                            comment,
                            course_id=course_id,
                            assignment_id=assignment.q_id,
                            group_id=group_id,
                            submission_id=submission.q_id,
                        )

                        for attachment in comment.attachments:
                            await self._converter.convert_and_store_attachment(
                                attachment,
                                course_id=course_id,
                                parent_api_object=comment,
                            )

    @staticmethod
    def _flatten_terms(bundles: list[CourseBundle]) -> list[gql.Term]:
        term_id_map: dict[str, gql.Term] = {}

        for bundle in bundles:
            term = bundle.course.term
            term_id_map[term.q_id] = term

        return list(term_id_map.values())

    async def _store_pages(self, pages: list[PageWithContent]):
        for page in pages:
            await self._converter.convert_and_store_page(page)

    async def _store_messages(
        self, messages: list[CourseMailItem | Announcement], known_course_ids: list[str]
    ):
        for message in messages:
            if isinstance(message, CourseMailItem):
                # Canvas includes mail for information "courses", but they're not visible in the allCourses graphql item.
                # Since they are not indexed anywhere else, we will discard them here
                if message.course_id in known_course_ids:
                    await self._converter.convert_and_store_mail_item(message)
                else:
                    _logger.debug(
                        'Discarding mail item "%s" (id="%s") because it\'s not in the list of courses for this sync (id="%s")',
                        message.subject,
                        message.id,
                        message.course_id,
                    )
            elif (
                isinstance(message, Announcement)
                and message.course_id in known_course_ids
            ):
                await self._converter.convert_and_store_announcement(message)

                for attachment in message.attachments:
                    await self._converter.convert_and_store_attachment(
                        attachment=attachment,
                        course_id=message.course_id,
                        parent_api_object=message,
                    )



import logging
from datetime import datetime
from typing import Any

import libqcanvas_clients.canvas as canvas
from sqlalchemy.ext.asyncio import AsyncSession

import libqcanvas.database.tables as db
import libqcanvas.gql_queries as gql
from libqcanvas.database import resource_links
from libqcanvas.database.tables import ModulePageType
from libqcanvas.net.canvas import CourseMailItem, PageWithContent
from libqcanvas.net.resources.extracting import CanvasFileExtractor
from libqcanvas.net.resources.extracting.link_extractor import LinkExtractor
from libqcanvas.net.sync.canvas_sync_observer import CanvasSyncObservable
from libqcanvas.util import (
    remove_screenreader_elements,
    remove_unwanted_whitespaces,
)

_logger = logging.getLogger(__name__)

NO_TITLE = "No Title"
NO_AUTHOR = "Unknown Author"


class APIDataImporter(CanvasSyncObservable):
    """
    An APIDataImporter accepts canvas API objects and converts and adds them to the database
    """

    _type_map = {
        gql.ShallowCourse: db.Course,
        gql.Term: db.Term,
        gql.Module: db.Module,
        gql.Assignment: db.Assignment,
        gql.AssignmentGroup: db.AssignmentGroup,
        gql.File: db.Page,
        gql.Submission: db.AssignmentSubmission,
        gql.SubmissionComment: db.SubmissionComment,
        PageWithContent: db.Page,
        CourseMailItem: db.Message,
        canvas.Announcement: db.Message,
    }

    def __init__(self, session: AsyncSession):
        super().__init__()
        self._session = session

    async def convert_and_store_course(
        self, course: gql.ShallowCourse, *, panopto_folder_id: str | None
    ) -> None:
        db_course: db.Course = await self._find_or_create_db_entry(course)
        is_new = _is_record_new(db_course)

        db_course.id = course.q_id
        db_course.name = remove_unwanted_whitespaces(course.name) or NO_TITLE
        db_course.term_id = course.term.q_id

        if panopto_folder_id is not None:
            db_course.panopto_folder_id = panopto_folder_id

        if is_new:
            self.notify_observers_for_updated_item(db_course)

    async def convert_and_store_module(
        self, module: gql.Module, *, course_id: str, position: int
    ) -> None:
        db_module: db.Module = await self._find_or_create_db_entry(module)

        is_new = _is_record_new(db_module)

        db_module.id = module.q_id
        db_module.name = module.name or NO_TITLE
        db_module.course_id = course_id
        db_module.position = position

        if is_new:
            self.notify_observers_for_updated_item(db_module)

    async def convert_and_store_page(self, page: PageWithContent) -> None:
        db_page: db.Page = await self._find_or_create_db_entry(page)

        if not (_is_record_new_or_updated(db_object=db_page, api_object=page)):
            # Updating the record is not going to do anything
            return

        db_page.id = page.q_id
        db_page.course_id = page.course.q_id
        db_page.module_id = page.module.q_id
        db_page.name = remove_unwanted_whitespaces(page.name) or NO_TITLE
        db_page.body = _clean_html_body(page.content)
        db_page.can_view = not page.is_locked
        db_page.unlock_at = page.unlock_at
        db_page.lock_at = page.lock_at
        db_page.creation_date = page.created_at
        db_page.last_modification_date = page.updated_at
        db_page.position = page.position
        db_page.page_type = ModulePageType.PAGE

        self.notify_observers_for_updated_item(db_page)

    async def convert_and_store_term(self, term: gql.Term) -> None:
        db_term: db.Term = await self._find_or_create_db_entry(term)

        is_new = _is_record_new(db_term)

        db_term.id = term.q_id
        db_term.end_date = term.end_at
        db_term.start_date = term.start_at
        db_term.name = remove_unwanted_whitespaces(term.name) or NO_TITLE

        if is_new:
            self.notify_observers_for_updated_item(db_term)

    async def convert_and_store_assignment_group(
        self, assignment_group: gql.AssignmentGroup, *, course_id: str, position: int
    ) -> None:
        db_group: db.AssignmentGroup = await self._find_or_create_db_entry(
            assignment_group
        )

        is_new = _is_record_new(db_group)

        db_group.id = assignment_group.q_id
        db_group.name = remove_unwanted_whitespaces(assignment_group.name) or NO_TITLE
        db_group.course_id = course_id
        db_group.group_weight = assignment_group.group_weight
        db_group.position = position

        if is_new:
            self.notify_observers_for_updated_item(db_group)

    async def convert_and_store_assignment(
        self, assignment: gql.Assignment, *, group_id: str
    ) -> None:
        db_assignment: db.Assignment = await self._find_or_create_db_entry(assignment)

        # Canvas creates bogus modtime updates, this tracks whether something was actually updated
        actually_updated = False
        can_view = assignment.lock_info.can_view or not assignment.lock_info.is_locked

        # check if the assignment has been unlocked since we last saved it
        if not _is_record_new(db_assignment):
            was_unlocked = not db_assignment.can_view and can_view
            actually_updated = was_unlocked

            if was_unlocked:
                _logger.debug(
                    "Assignment '%s' (id=%s) was unlocked",
                    assignment.name,
                    assignment.q_id,
                )
            elif not _is_api_object_newer(
                db_object=db_assignment, api_object=assignment
            ):
                # Updating the record is not going to do anything
                return

        _logger.debug(
            "Assignment %s (id=%s) was updated according to canvas",
            assignment.name,
            assignment.q_id,
        )

        # "boring" fields that don't really matter to the user if they're updated
        db_assignment.id = assignment.q_id
        db_assignment.course_id = assignment.course_id
        db_assignment.creation_date = assignment.created_at
        db_assignment.position = assignment.position
        db_assignment.group_id = group_id
        db_assignment.last_modification_date = assignment.updated_at

        db_assignment.can_view = can_view
        db_assignment.unlock_at = assignment.lock_info.unlock_at
        db_assignment.lock_at = assignment.lock_info.lock_at

        new_name = remove_unwanted_whitespaces(assignment.name) or NO_TITLE
        new_body = _clean_html_body(assignment.description)
        new_due_date = assignment.due_at
        new_max_score = assignment.points_possible

        # check if these fields are actually updated. Canvas seems to create bogus updates for assignments, possibly from
        # markers doing something in the background.

        if new_name != db_assignment.name:
            db_assignment.name = new_name
            actually_updated = True
            _logger.debug(
                "assignment name updated (name=%s, id=%s)",
                assignment.name,
                assignment.q_id,
            )

        if new_body != db_assignment.body:
            db_assignment.body = new_body
            actually_updated = True
            _logger.debug(
                "assignment body updated (name=%s, id=%s)",
                assignment.name,
                assignment.q_id,
            )

        if new_due_date != db_assignment.due_date:
            db_assignment.due_date = new_due_date
            actually_updated = True
            _logger.debug(
                "assignment due date updated (name=%s, id=%s)",
                assignment.name,
                assignment.q_id,
            )

        if new_max_score != db_assignment.max_score:
            db_assignment.max_score = new_max_score
            actually_updated = True
            _logger.debug(
                "assignment max score updated (name=%s, id=%s)",
                assignment.name,
                assignment.q_id,
            )

        if actually_updated:
            self.notify_observers_for_updated_item(db_assignment)
        else:
            _logger.debug(
                "Assignment %s (id=%s) wasn't actually modified",
                assignment.name,
                assignment.q_id,
            )

    async def convert_and_store_assignment_submission(
        self,
        submission: gql.Submission,
        *,
        assignment_id: str,
        group_id: str,
        course_id: str,
    ) -> None:
        db_submission: db.AssignmentSubmission = await self._find_or_create_db_entry(
            submission
        )

        if not _is_record_new_or_updated(db_submission, submission):
            return

        db_submission.id = submission.q_id

        db_submission.course_id = course_id
        db_submission.group_id = group_id
        db_submission.assignment_id = assignment_id

        db_submission.attempt = submission.attempt
        db_submission.last_modification_date = submission.updated_at
        db_submission.creation_date = submission.created_at
        db_submission.score = submission.score

        self.notify_observers_for_updated_item(db_submission)

    async def convert_and_store_submission_comment(
        self,
        comment: gql.SubmissionComment,
        *,
        assignment_id: str,
        group_id: str,
        course_id: str,
        submission_id: str,
    ):
        db_comment: db.SubmissionComment = await self._find_or_create_db_entry(comment)

        if not _is_record_new_or_updated(db_comment, comment):
            return

        db_comment.id = comment.q_id

        db_comment.course_id = course_id
        db_comment.group_id = group_id
        db_comment.assignment_id = assignment_id
        db_comment.submission_id = submission_id

        db_comment.body = comment.html_comment
        db_comment.author = comment.author.name
        db_comment.creation_date = comment.created_at
        db_comment.last_modification_date = comment.updated_at

        self.notify_observers_for_updated_item(db_comment)

    async def convert_and_store_announcement(
        self, announcement: canvas.Announcement
    ) -> None:
        db_message: db.Message = await self._find_or_create_db_entry(announcement)

        # Canvas doesn't seem to track last modification date on announcements for some reason
        # Otherwise, I would check for modification time here
        is_new = _is_record_new(db_message)

        db_message.id = str(announcement.id)
        db_message.course_id = announcement.course_id
        db_message.creation_date = announcement.created_at
        db_message.name = remove_unwanted_whitespaces(announcement.title) or NO_TITLE
        db_message.body = _clean_html_body(announcement.message)
        db_message.sender_name = announcement.user_name or NO_AUTHOR
        db_message.has_been_read = False

        # I don't think this will have any effect since the announcement just disappears when it's locked.
        db_message.can_view = announcement.locked_for_user
        db_message.lock_at = announcement.lock_at

        if is_new:
            self.notify_observers_for_updated_item(db_message)

    async def convert_and_store_attachment(
        self,
        attachment: canvas.RemoteFile | gql.Attachment,
        *,
        course_id: str,
        parent_api_object: Any,
    ) -> None:
        resource = _create_resource_from_attachment(attachment, course_id=course_id)

        if await self._add_resource_if_missing(resource):
            db_type = self._type_map[type(parent_api_object)]
            content_id = _id_from_api_object(parent_api_object)

            await resource_links.create_resource_link_a(
                self._session,
                content_type=db_type,
                content_id=content_id,
                resource=resource,
                link_type="attachment",
            )
            self.notify_observers_for_updated_item(resource)

    async def convert_and_store_mail_item(self, mail: CourseMailItem) -> None:
        db_message: db.Message = await self._find_or_create_db_entry(mail)

        is_new = _is_record_new(db_message)

        db_message.id = mail.id
        db_message.course_id = mail.course_id
        db_message.creation_date = mail.date
        db_message.name = remove_unwanted_whitespaces(mail.subject) or NO_TITLE
        db_message.body = _convert_plaintext_to_html(
            remove_unwanted_whitespaces(mail.body)
        )
        db_message.sender_name = mail.author_name or NO_AUTHOR
        db_message.has_been_read = False
        db_message.can_view = True  # I don't think mail can be locked

        if is_new:
            self.notify_observers_for_updated_item(db_message)

    async def _find_or_create_db_entry(self, obj: Any):
        obj_type = type(obj)

        if obj_type not in self._type_map.keys():
            raise TypeError(f"{obj_type} is not present in _type_map")

        obj_id = _id_from_api_object(obj)
        db_type = self._type_map[obj_type]
        db_object = await self._session.get(db_type, obj_id)

        if db_object is None:
            _logger.debug(
                'Converting new %s (api) to %s (db) (id="%s")',
                obj_type.__name__,
                db_type.__name__,
                obj_id,
            )
            db_object = db_type()
            self._session.add(db_object)
        else:
            _logger.debug(
                'Found existing %s (db) for %s (api) (id="%s")',
                db_type.__name__,
                obj_type.__name__,
                obj_id,
            )

        return db_object

    async def _add_resource_if_missing(self, resource: db.Resource) -> bool:
        does_not_exist = (await self._session.get(db.Resource, resource.id)) is None

        if does_not_exist:
            _logger.debug("New resource %s (id=%s)", resource.file_name, resource.id)
            self._session.add(resource)
            self.notify_observers_for_updated_item(resource)

        return does_not_exist


def _id_from_api_object(obj: Any) -> str | int | tuple[str, str]:
    # Sanity check
    if hasattr(obj, "q_id") and hasattr(obj, "id"):
        _logger.warning(f"!?!?!? {obj} has `q_id` AND `id`???")

    if hasattr(obj, "q_id"):
        return obj.q_id
    elif hasattr(obj, "id"):
        return obj.id
    else:
        raise TypeError(f"{type(obj)} can not be handled")


def _is_record_new_or_updated(
    db_object: object | db.ModificationDate, api_object: object
) -> bool:
    if _is_record_new(db_object):
        return True
    else:
        return _is_api_object_newer(db_object, api_object)


def _is_api_object_newer(
    db_object: object | db.ModificationDate, api_object: object
) -> bool:
    if (
        hasattr(api_object, "updated_at")
        and isinstance(api_object.updated_at, datetime)
        and isinstance(db_object.last_modification_date, datetime)
    ):
        return db_object.last_modification_date < api_object.updated_at
    else:
        return False


def _is_record_new(db_obj: object) -> bool:
    if hasattr(db_obj, "id"):
        return db_obj.id is None
    else:
        return False


def _clean_html_body(page: str) -> str:
    return remove_screenreader_elements(remove_unwanted_whitespaces(page))


def _convert_plaintext_to_html(plaintext: str) -> str:
    return plaintext.replace("\n", "<br/>\n")


def _create_resource_from_attachment(
    attachment: gql.Attachment | canvas.RemoteFile, *, course_id: str
) -> db.Resource:
    if hasattr(attachment, "q_id"):
        attachment_id = attachment.q_id
    else:
        attachment_id = attachment.id

    return db.Resource(
        id=f"{CanvasFileExtractor.extractor_id}{LinkExtractor.RESOURCE_ID_SEPARATOR}{attachment_id}",
        url=attachment.url,
        course_id=course_id,
        discovery_date=datetime.now().astimezone(),
        file_name=attachment.display_name,
    )



from pathlib import Path
from typing import AsyncIterator, NamedTuple

from bs4 import Tag
from httpx import URL
from libqcanvas_clients.canvas import CanvasClient

import libqcanvas.database.tables as db
from libqcanvas.net.resources.download.download_progress import DownloadProgress
from libqcanvas.net.resources.extracting.link_extractor import LinkExtractor


class _CanvasResourceInfo(NamedTuple):
    file_id: str
    course_id: str


class CanvasFileExtractor(LinkExtractor):
    extractor_id = "canvas"

    def __init__(self, canvas_client: CanvasClient):
        super().__init__(tag_whitelist=["a", "img"])
        self._canvas_client = canvas_client

    def _is_tag_valid(self, tag: Tag) -> bool:
        if "data-api-returntype" not in tag.attrs:
            return False
        elif tag.attrs["data-api-returntype"] != "File":
            return False

        return True

    async def download(
        self, resource: db.Resource, destination: Path
    ) -> AsyncIterator[DownloadProgress]:
        yield DownloadProgress(0, 0)

        response = await self._canvas_client.get_file_download_stream(resource.url)

        try:
            async for progress in self._download_stream_response(
                destination, response, resource.file_size
            ):
                yield progress
        finally:
            await response.aclose()

    async def _extract_resource(self, tag: Tag) -> db.Resource:
        file_id, course_id = self._course_and_file_id_from_tag(tag)
        file = await self._canvas_client.get_file(file_id=file_id, course_id=course_id)

        return db.Resource(
            url=file.url,
            file_name=file.display_name,
            file_size=file.size,
        )

    def _tag_id(self, tag: Tag) -> str:
        return self._course_and_file_id_from_tag(tag).file_id

    def _course_and_file_id_from_tag(self, tag: Tag) -> _CanvasResourceInfo:
        # https://canvas_instance/api/v1/courses/123/files/456...
        # ---------- Extract these parts:        ^^^  and  ^^^

        parts = self._tag_api_endpoint(tag).path.rsplit("/", 3)
        return _CanvasResourceInfo(file_id=parts[-1], course_id=parts[-3])

    @staticmethod
    def _tag_api_endpoint(tag: Tag) -> URL:
        return URL(tag.attrs["data-api-endpoint"])



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from libqcanvas.gql_queries import Module, ShallowCourse


@dataclass
class PageWithContent:
    q_id: str
    updated_at: Optional[datetime]
    created_at: Optional[datetime]
    module: Module
    course: ShallowCourse
    position: int
    name: Optional[str] = None
    content: Optional[str] = None
    is_locked: bool = False
    unlock_at: Optional[datetime] = None
    lock_at: Optional[datetime] = None



import logging
from datetime import datetime

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship

from libqcanvas.database.tables.base import Base
from libqcanvas.database.tables.id import ID
from libqcanvas.database.tables.resource_download_state import ResourceDownloadState

_logger = logging.getLogger(__name__)


# noinspection PyDataclass
class Resource(Base, ID):
    __tablename__ = "resources"

    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))
    course = relationship("Course", back_populates="resources")

    url: Mapped[str]
    file_name: Mapped[str]
    # default_factory won't work here because init=False
    discovery_date: Mapped[datetime]
    file_size: Mapped[int | None]
    download_state: Mapped[ResourceDownloadState] = mapped_column(
        default=ResourceDownloadState.NOT_DOWNLOADED
    )
    download_error_message: Mapped[str | None]

    polymorphic_type: Mapped[str]

    __mapper_args__ = {
        "polymorphic_on": "polymorphic_type",
        "polymorphic_identity": "resource",
    }


# noinspection PyDataclass
class PanoptoResource(Resource):
    __tablename__ = "panopto_resources"

    id: Mapped[str] = mapped_column(ForeignKey("resources.id"), primary_key=True)

    duration_seconds: Mapped[int]
    recording_date: Mapped[datetime]

    # Panopto/Canvas have this stupid "custom_context_delivery" which is a pain in the ass because it has nothing to do
    # with the actual ID of the video. In this case, this object's id may not be useful in any way (thanks painopto),
    # but delivery_id will always be the true ID of the video.
    # !!!! It has been observed that different "custom_context_delivery"s CAN link to the same video !!!!
    delivery_id: Mapped[str]

    __mapper_args__ = {"polymorphic_identity": "panopto_resource"}



import html
from datetime import datetime
from pathlib import Path
from typing import AsyncIterator

from bs4 import Tag
from httpx import URL
from libqcanvas_clients.canvas import CanvasClient
from libqcanvas_clients.panopto import PanoptoClient

import libqcanvas.database.tables as db
from libqcanvas.net.panopto import IFramePanoptoVideoExtractor, video_downloader
from libqcanvas.net.resources.download.download_progress import DownloadProgress
from libqcanvas.net.resources.extracting.link_extractor import LinkExtractor


class EmbeddedPanoptoExtractor(LinkExtractor):
    extractor_id = "panopto_embedded"

    def __init__(self, canvas_client: CanvasClient, panopto_client: PanoptoClient):
        super().__init__(
            tag_whitelist=["iframe"],
            is_video_extractor=True,
        )
        self._embedded_video_info_extractor = IFramePanoptoVideoExtractor(canvas_client)
        self._panopto_client = panopto_client

    async def download(
        self, resource: db.Resource, destination: Path
    ) -> AsyncIterator[DownloadProgress]:
        if not isinstance(resource, db.PanoptoResource):
            raise TypeError()

        async for progress in video_downloader.download_video(
            video_id=resource.delivery_id,
            client=self._panopto_client,
            destination=destination,
        ):
            yield progress

    def _is_tag_valid(self, tag: Tag) -> bool:
        if "src" not in tag.attrs:
            return False

        src_raw = tag.attrs["src"]
        src = URL(src_raw)

        if self._is_iframe_embedded_panopto_video(
            src
        ) or self._is_resource_link_lookup_lti_iframe(src_raw):
            return True
        elif self._is_custom_context_lti_iframe(src_raw):
            lti_url = URL(src.params["url"])
            return self._is_custom_context_lti_url(lti_url)
        else:
            return False

    def _tag_id(self, tag: Tag) -> str:
        src_raw = tag.attrs["src"]
        src = URL(html.unescape(src_raw))

        # This might result in resource duplication if somehow there are 2 instances of the same video embedded with
        # both methods, but I REALLY don't care anymore and I want something that at least works.
        # I call it PAINopto for a reason.
        if self._is_iframe_embedded_panopto_video(src):
            return src.params["id"]
        elif self._is_custom_context_lti_iframe(src_raw):
            lti_url = URL(src.params["url"])
            return lti_url.params["custom_context_delivery"]
        elif self._is_resource_link_lookup_lti_iframe(src_raw):
            return src.params["resource_link_lookup_uuid"]
        else:
            raise ValueError(f"Can't tell what type of video it is from the URL: {src}")

    async def _extract_resource(self, tag: Tag) -> db.Resource:
        src_raw = tag.attrs["src"]
        src = URL(src_raw)

        if self._is_iframe_embedded_panopto_video(src):
            delivery_id = src.params["id"]
            video_url = str(src)
        else:
            info = (
                await self._embedded_video_info_extractor.retrieve_embedded_video_info(
                    src
                )
            )
            delivery_id = info.delivery_id
            video_url = info.viewer_link

        delivery_info = await self._panopto_client.get_session_info(delivery_id)

        return db.PanoptoResource(
            url=video_url,
            delivery_id=delivery_id,
            file_name=delivery_info.name
            + ".mp4",  # All panopto videos are downloaded as mp4s, but doesn't include the extension in the name
            duration_seconds=delivery_info.duration,
            recording_date=datetime.fromisoformat(delivery_info.start_time),
        )

    @staticmethod
    def _is_iframe_embedded_panopto_video(src: URL):
        # Check for urls that look like this: https://instance.panopto.com/Panopto/Pages/Embed.aspx?id=123456-1234-124556...
        #                                                                                ^^^^^^^^^^
        # Also check that "id" parameter is supplied
        return src.path.endswith("Embed.aspx") and "id" in src.params

    @staticmethod
    def _is_custom_context_lti_iframe(raw_src: str):
        # The URL is html encoded because it is from an iframe
        src = URL(html.unescape(raw_src))

        def _remove_empty(string: str) -> bool:
            return len(string) > 1

        path = list(filter(_remove_empty, src.path.split("/", 4)))

        # Check the URL path https://canvas_instance/courses/12345/external_tools?url=whatever.panopto.com/?custom_context_delivery=... (encoded)
        #                                                          ^^^^^^^^^^^^^^
        # Also check that "url" parameter is supplied
        return len(path) >= 3 and path[2] == "external_tools" and "url" in src.params

    @staticmethod
    def _is_resource_link_lookup_lti_iframe(raw_src: str) -> bool:
        # The URL is html encoded because it is from an iframe
        src = URL(html.unescape(raw_src))

        def _remove_empty(string: str) -> bool:
            return len(string) > 1

        path = list(filter(_remove_empty, src.path.split("/", 4)))

        # Check the URL path https://canvas_instance/courses/12345/external_tools/retrieve?resource_link_lookup_uuid=...
        #                                                          ^^^^^^^^^^^^^^
        # Also check that "resource_link_lookup_uuid" parameter is supplied
        return (
            len(path) >= 3
            and path[2] == "external_tools"
            and "resource_link_lookup_uuid" in src.params
        )

    @staticmethod
    def _is_custom_context_lti_url(lti_url: URL) -> bool:
        # Check for urls that look like this:
        # https://instance.panopto.com/Panopto/LTI/LTI.aspx?custom_context_delivery=12345-1234567-13454...
        #                  ^^^^^^^^^^^
        # Also check that "custom_context_delivery" parameter is supplied
        return (
            "panopto.com" in lti_url.host
            and "custom_context_delivery" in lti_url.params
        )



"""
Generated by qenerate plugin=pydantic_v2. DO NOT MODIFY MANUALLY!
"""

from collections.abc import Callable  # noqa: F401 # pylint: disable=W0611
from datetime import datetime  # noqa: F401 # pylint: disable=W0611
from enum import Enum  # noqa: F401 # pylint: disable=W0611
from typing import (  # noqa: F401 # pylint: disable=W0611
    Any,
    Optional,
    Union,
)

from pydantic import (  # noqa: F401 # pylint: disable=W0611
    BaseModel,
    Extra,
    Field,
    Json,
)

from libqcanvas.gql_queries.attachment import Attachment

DEFINITION = """
fragment Attachment on File {
    id
    url
    displayName
    mimeClass
    contentType
    createdAt
    updatedAt
}

query AssignmentIndex($group_id: ID!) {
    node(id: $group_id) {
        ... on AssignmentGroup {
            assignmentsConnection {
                nodes {
                    id
                    name
                    description
                    dueAt
                    position
                    pointsPossible
                    courseId
                    assignmentGroupId
                    createdAt
                    updatedAt

                    lockInfo {
                        unlockAt
                        lockAt
                        canView
                        isLocked
                    }

                    submissionsConnection {
                        nodes {
                            id
                            score
                            attempt
                            body
                            createdAt
                            updatedAt
                            groupId

                            attachments {
                                ...Attachment
                            }

                            commentsConnection {
                                nodes {
                                    id
                                    htmlComment
                                    attachments {
                                        ...Attachment
                                    }
                                    createdAt
                                    updatedAt

                                    author {
                                        name
                                    }
                                }
                            }
                        }
                    }
                }

            }
        }
    }
}
"""


class ConfiguredBaseModel(BaseModel):
    class Config:
        extra = "forbid"


class Node(ConfiguredBaseModel): ...


class LockInfo(ConfiguredBaseModel):
    unlock_at: Optional[datetime] = Field(..., alias="unlockAt")
    lock_at: Optional[datetime] = Field(..., alias="lockAt")
    can_view: Optional[bool] = Field(..., alias="canView")
    is_locked: bool = Field(..., alias="isLocked")


class User(ConfiguredBaseModel):
    name: Optional[str] = Field(..., alias="name")


class SubmissionComment(ConfiguredBaseModel):
    q_id: str = Field(..., alias="id")
    html_comment: Optional[str] = Field(..., alias="htmlComment")
    attachments: list[Attachment] = Field(..., alias="attachments")
    created_at: datetime = Field(..., alias="createdAt")
    updated_at: Optional[datetime] = Field(..., alias="updatedAt")
    author: Optional[User] = Field(..., alias="author")


class SubmissionCommentConnection(ConfiguredBaseModel):
    nodes: Optional[list[Optional[SubmissionComment]]] = Field(..., alias="nodes")


class Submission(ConfiguredBaseModel):
    q_id: str = Field(..., alias="id")
    score: Optional[float] = Field(..., alias="score")
    attempt: int = Field(..., alias="attempt")
    body: Optional[str] = Field(..., alias="body")
    created_at: Optional[datetime] = Field(..., alias="createdAt")
    updated_at: Optional[datetime] = Field(..., alias="updatedAt")
    group_id: Optional[str] = Field(..., alias="groupId")
    attachments: Optional[list[Attachment]] = Field(..., alias="attachments")
    comments_connection: SubmissionCommentConnection = Field(
        ..., alias="commentsConnection"
    )


class SubmissionConnection(ConfiguredBaseModel):
    nodes: Optional[list[Optional[Submission]]] = Field(..., alias="nodes")


class Assignment(ConfiguredBaseModel):
    q_id: str = Field(..., alias="id")
    name: Optional[str] = Field(..., alias="name")
    description: Optional[str] = Field(..., alias="description")
    due_at: Optional[datetime] = Field(..., alias="dueAt")
    position: Optional[int] = Field(..., alias="position")
    points_possible: Optional[float] = Field(..., alias="pointsPossible")
    course_id: Optional[str] = Field(..., alias="courseId")
    assignment_group_id: Optional[str] = Field(..., alias="assignmentGroupId")
    created_at: Optional[datetime] = Field(..., alias="createdAt")
    updated_at: Optional[datetime] = Field(..., alias="updatedAt")
    lock_info: Optional[LockInfo] = Field(..., alias="lockInfo")
    submissions_connection: Optional[SubmissionConnection] = Field(
        ..., alias="submissionsConnection"
    )


class AssignmentConnection(ConfiguredBaseModel):
    nodes: Optional[list[Optional[Assignment]]] = Field(..., alias="nodes")


class AssignmentGroup(Node):
    assignments_connection: AssignmentConnection = Field(
        ..., alias="assignmentsConnection"
    )


class AssignmentIndexQueryData(ConfiguredBaseModel):
    node: Optional[Union[AssignmentGroup, Node]] = Field(..., alias="node")


def query(query_func: Callable, **kwargs: Any) -> AssignmentIndexQueryData:
    """
    This is a convenience function which queries and parses the data into
    concrete types. It should be compatible with most GQL clients.
    You do not have to use it to consume the generated data classes.
    Alternatively, you can also mime and alternate the behavior
    of this function in the caller.

    Parameters:
        query_func (Callable): Function which queries your GQL Server
        kwargs: optional arguments that will be passed to the query function

    Returns:
        AssignmentIndexQueryData: queried data parsed into generated classes
    """
    raw_data: dict[Any, Any] = query_func(DEFINITION, **kwargs)
    return AssignmentIndexQueryData(**raw_data)



"""
Generated by qenerate plugin=pydantic_v2. DO NOT MODIFY MANUALLY!
"""

from collections.abc import Callable  # noqa: F401 # pylint: disable=W0611
from datetime import datetime  # noqa: F401 # pylint: disable=W0611
from enum import Enum  # noqa: F401 # pylint: disable=W0611
from typing import (  # noqa: F401 # pylint: disable=W0611
    Any,
    Optional,
    Union,
)

from pydantic import (  # noqa: F401 # pylint: disable=W0611
    BaseModel,
    Extra,
    Field,
    Json,
)

DEFINITION = """
query CourseIndex {
    allCourses {
        _id
        name
        courseNickname
        settings {
            # This value isn't actually used, but something needs to be included to determine if the settings object is null
            restrictStudentPastView
        }
        term {
            id
            startAt
            endAt
            name
        }
    }
}
"""


class ConfiguredBaseModel(BaseModel):
    class Config:
        extra = "forbid"


class CourseSettings(ConfiguredBaseModel):
    restrict_student_past_view: Optional[bool] = Field(
        ..., alias="restrictStudentPastView"
    )


class Term(ConfiguredBaseModel):
    q_id: str = Field(..., alias="id")
    start_at: Optional[datetime] = Field(..., alias="startAt")
    end_at: Optional[datetime] = Field(..., alias="endAt")
    name: Optional[str] = Field(..., alias="name")


class Course(ConfiguredBaseModel):
    q_id: str = Field(..., alias="_id")
    name: str = Field(..., alias="name")
    course_nickname: Optional[str] = Field(..., alias="courseNickname")
    settings: Optional[CourseSettings] = Field(..., alias="settings")
    term: Optional[Term] = Field(..., alias="term")


class CourseIndexQueryData(ConfiguredBaseModel):
    all_courses: Optional[list[Course]] = Field(..., alias="allCourses")


def query(query_func: Callable, **kwargs: Any) -> CourseIndexQueryData:
    """
    This is a convenience function which queries and parses the data into
    concrete types. It should be compatible with most GQL clients.
    You do not have to use it to consume the generated data classes.
    Alternatively, you can also mime and alternate the behavior
    of this function in the caller.

    Parameters:
        query_func (Callable): Function which queries your GQL Server
        kwargs: optional arguments that will be passed to the query function

    Returns:
        CourseIndexQueryData: queried data parsed into generated classes
    """
    raw_data: dict[Any, Any] = query_func(DEFINITION, **kwargs)
    return CourseIndexQueryData(**raw_data)



import logging
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Self, override

import libqcanvas.database.tables as db
from libqcanvas.net.sync.canvas_sync_observer import CanvasSyncObserver

_logger = logging.getLogger(__name__)


@dataclass
class CourseUpdates:
    updates: dict[type, set[str | int]] = field(
        default_factory=lambda: defaultdict(set)
    )

    def new_content_found(self, content: object) -> None:
        if isinstance(content, db.Module):
            self.updates[db.Module].add(content.id)
        elif isinstance(content, db.Page):
            self.updates[db.Page].add(content.id)
            self.updates[db.Module].add(content.module_id)
        elif isinstance(content, db.Message):
            self.updates[db.Message].add(content.id)
        elif isinstance(content, db.AssignmentGroup):
            self.updates[db.AssignmentGroup].add(content.id)
        elif isinstance(content, db.Assignment):
            self.updates[db.Assignment].add(content.id)
            self.updates[db.AssignmentGroup].add(content.group_id)
        elif isinstance(content, db.AssignmentSubmission):
            self.updates[db.AssignmentSubmission].add(content.id)
            self.updates[db.AssignmentGroup].add(content.group_id)
            self.updates[db.Assignment].add(content.assignment_id)
        elif isinstance(content, db.SubmissionComment):
            self.updates[db.SubmissionComment].add(content.id)
            self.updates[db.AssignmentSubmission].add(content.submission_id)
            self.updates[db.AssignmentGroup].add(content.group_id)
            self.updates[db.Assignment].add(content.assignment_id)
        elif isinstance(content, db.Resource):
            self.updates[db.Resource].add(content.id)
        else:
            _logger.warning(
                "Content of type %s was not properly handled by the sync receipt tracker!",
                type(content).__name__,
            )

    def was_updated(self, content: object) -> bool:
        content_type = type(content)

        if hasattr(content, "id"):
            if content_type in self.updates:
                return content.id in self.updates[content_type]
        else:
            _logger.warning(
                "Object without `id` attr passed to `was_updated` (type=%s)",
                content_type.__name__,
            )

        return False

    def finalise(self) -> None:
        """
        Finalises the updates list.
        Converts the `updates` `defaultdict` into a regular dictionary to make it harder to accidentally add something to it
        """
        self.updates = dict(self.updates)

    def __getitem__(self, item: type) -> set[str | int] | None:
        if not isinstance(item, type):
            raise TypeError("item")

        return self.updates.get(item, None)


@dataclass
class SyncReceipt(CanvasSyncObserver, CourseUpdates):
    updates_by_course: dict[str, CourseUpdates] = field(
        default_factory=lambda: defaultdict(CourseUpdates)
    )

    def new_content_found(self, content: object) -> None:
        if isinstance(content, db.Course):
            self.updates[db.Course].add(content.id)
        else:
            # Need to specify CourseUpdates because of the diamond problem
            CourseUpdates.new_content_found(self, content)

            if hasattr(content, "course_id"):
                if content.course_id not in self.updates_by_course:
                    self.updates[db.Course].add(content.course_id)

                self.updates_by_course[content.course_id].new_content_found(content)

    @override
    def finalise(self) -> Self:
        """
        Finalises the sync receipt.
        Converts all `defaultdict`s to regular dictionaries to make it harder to accidentally add things to them
        """
        self.updates_by_course = dict(self.updates_by_course)

        for value in self.updates_by_course.values():
            value.finalise()

        return self


_empty = SyncReceipt()
_empty.finalise()


def empty_receipt() -> SyncReceipt:
    """
    An empty sync receipt instance for calling functions that would normally require one from an antecedent sync operation
    """
    global _empty
    return _empty
